"""
Contains relatively high-level tests of Simulation object, testing a variety of simulation parameter combinations
to assert simulation outputs are as expected.
"""
import logging
import os
import sys
import unittest
from configparser import ConfigParser

import numpy as np

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

try:
    from io import StringIO
except ImportError as ie:  # Python 2.x support
    from cStringIO import StringIO

from pycoalescence.simulation import Simulation
from pycoalescence.coalescence_tree import CoalescenceTree
from pycoalescence import __version__ as pycoalescence_version
from pycoalescence.map import Map
from setup_tests import setUpAll, tearDownAll, skipLongTest


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestFileCreation(unittest.TestCase):
    """
    Tests the main Simulation set up routine by running some tiny simulations and checking that simulation parameters
    are passed properly.
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=logging.CRITICAL)
        self.coal.set_simulation_parameters(
            10,
            38,
            "output/test_output/test_output2/",
            0.1,
            4,
            4,
            deme=1,
            sample_size=1.0,
            max_time=2,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        self.coal.set_map_parameters(
            "null",
            10,
            10,
            "null",
            10,
            10,
            0,
            0,
            "null",
            20,
            20,
            0,
            0,
            1,
            "null",
            "null",
        )
        self.coal.set_speciation_rates([0.1, 0.2])
        self.coal.run()

    def testFileCreation(self):
        """
        Checks that outputting is to the correct place and folder structure is created properly.
        """
        self.assertTrue(os.path.isfile(self.coal.output_database))
        self.assertEqual(
            os.path.join(self.coal.output_database),
            os.path.join(
                self.coal.output_directory,
                "data_{}_{}.db".format(self.coal.task, self.coal.seed),
            ),
        )


class TestFileNaming(unittest.TestCase):
    """
    Tests that the file naming structure makes sense
    """

    def testNoneNaming(self):
        """
        Tests that the fine map file naming throws the correct error when called 'none'.
        :return:
        """
        coal = Simulation()
        coal.set_simulation_parameters(
            100000,
            10000,
            "output",
            0.1,
            4,
            4,
            deme=1,
            sample_size=1.0,
            max_time=2,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        with self.assertRaises(ValueError):
            coal.set_map_files(sample_file="null", fine_file="null")


class TestSetSeed(unittest.TestCase):
    """Tests that seeds are correctly set, and errors are thrown where appropriate."""

    def testIncorrectSeedRaisesErrors(self):
        """Tests that an error is raised if the seed is too large."""
        s = Simulation()
        with self.assertRaises(ValueError):
            s.set_seed(2147483647)
        with self.assertRaises(ValueError):
            s.set_simulation_parameters(
                seed=2147483647,
                task=1,
                output_directory="output",
                min_speciation_rate=0.1,
            )

    def testBasicSeedSetting(self):
        """Tests that the basic seed setting works as intended."""
        s = Simulation()
        s.set_seed(1)
        self.assertEqual(1, s.seed)
        s = Simulation()
        s.set_simulation_parameters(
            seed=1,
            task=1,
            output_directory="output",
            min_speciation_rate=0.1,
            spatial=False,
        )
        self.assertEqual(1, s.seed)

    def testModifiedSeedSetting(self):
        """Tests that the modification to seed setting works as intended."""
        s = Simulation(logging_level=60)
        s.set_seed(0)
        self.assertEqual(1073741823, s.seed)
        s = Simulation(logging_level=60)
        s.set_simulation_parameters(
            seed=0,
            task=1,
            output_directory="output",
            min_speciation_rate=0.1,
            spatial=False,
        )
        self.assertEqual(1073741823, s.seed)
        s = Simulation(logging_level=60)
        s.set_seed(-1)
        self.assertEqual(1073741824, s.seed)
        s = Simulation(logging_level=60)
        s.set_seed(-2)
        self.assertEqual(1073741825, s.seed)
        s = Simulation(logging_level=60)
        s.set_seed(-300)
        self.assertEqual(1073742123, s.seed)


class TestOffsetMismatch(unittest.TestCase):
    """
    Tests if simulations correctly detect when offsets between fine and sample, or fine and coarse maps do not make
    sense.
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the class by creating simulation object with the desired map structure.
        """
        cls.coal1 = Simulation(logging_level=logging.CRITICAL)
        cls.coal1.set_simulation_parameters(
            seed=1,
            task=38,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=4,
            max_time=2,
            dispersal_relative_cost=1,
            min_num_species=1,
            cutoff=0.0,
        )
        cls.coal2 = Simulation()
        cls.coal2.set_simulation_parameters(
            seed=1,
            task=38,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=4,
            max_time=2,
            dispersal_relative_cost=1,
            min_num_species=1,
            cutoff=0.0,
        )

    def testRaisesErrorFineSampleOffset(self):
        """
        Tests the correct error is raised when offsetting is incorrect
        """
        with self.assertRaises(ValueError):
            self.coal1.set_map_files(
                sample_file="sample/SA_samplemaskINT.tif",
                fine_file="sample/SA_sample_fine_offset.tif",
            )

    def testRaisesErrorFineSampleOffset2(self):
        """
        Tests the correct error is raised when offsetting is incorrect
        """
        with self.assertRaises(ValueError):
            self.coal2.set_map_files(
                sample_file="null",
                fine_file="sample/SA_sample_fine_offset.tif",
                coarse_file="sample/SA_sample_coarse.tif",
            )

    def testRaisesErrorAfterIncorrectSamplegridBoundaries(self):
        """
        Checks that setting incorrect limits for the sample grid outside of the fine map causes an error to be thrown.
        """
        sim = Simulation()
        sim.set_simulation_parameters(
            seed=1,
            task=1,
            output_directory="output",
            min_speciation_rate=0.001,
            sigma=2,
        )
        sim.set_map_files(
            sample_file="null",
            fine_file="sample/SA_sample_fine_offset.tif",
            coarse_file="none",
        )
        sim.sample_map.x_offset = 10000
        sim.sample_map.y_offset = 10000
        sim.grid.x_size = 1000
        sim.grid.y_size = 1000
        sim.grid.file_name = "set"
        with self.assertRaises(ValueError):
            sim.finalise_setup()
            sim.run_coalescence()


class TestSimulationRaisesErrors(unittest.TestCase):
    """
    Tests that protracted and normal simulations raise the correct errors.
    """

    def testNormalRaisesError(self):
        """
        Tests a normal simulation raises an error when no files exist
        """
        c = Simulation()
        c.set_simulation_parameters(
            5,
            4,
            "output",
            0.1,
            4,
            4,
            1,
            0.01,
            2,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
        )
        c.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        # Now change map name to something that doesn't exist

        c.fine_map.file_name = "not_here.tif"
        with self.assertRaises(IOError):
            c.finalise_setup()
        with self.assertRaises(RuntimeError):
            c.run_coalescence()

    def testProtractedRaisesError(self):
        """
        Tests a protracted simulation raises an error if there is a problem
        """
        c = Simulation(logging_level=logging.ERROR)
        c.set_simulation_parameters(
            6,
            4,
            "output",
            0.1,
            4,
            4,
            1,
            0.01,
            2,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
            protracted=True,
        )
        c.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        # Now change map name to something that doesn't exist

        c.fine_map.file_name = "not_here.tif"
        with self.assertRaises(IOError):
            c.finalise_setup()
        with self.assertRaises(RuntimeError):
            c.run_coalescence()


class TestSimulationConfigReadWrite(unittest.TestCase):
    """
    Tests the reading and writing to a config text file.
    Independently tests the main config, map config and time config writing ability.
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=logging.CRITICAL)
        self.coal.set_simulation_parameters(
            1,
            23,
            "output",
            0.1,
            4,
            4,
            1,
            1.0,
            max_time=200,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
        )
        self.coal.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal.add_sample_time(0.0)
        self.coal.add_sample_time(1.0)
        self.coal.set_speciation_rates([0.1, 0.2])
        self.coal.create_config("output/conf1.txt")

    def testConfigWriteMain(self):
        """
        Tests that the main configuration file has been correctly generated.
        """
        with open("output/conf1.txt", "r") as mapconf:
            lines = mapconf.readlines()
            lines = [x.strip() for x in lines]
            self.assertEqual(lines[0], "[main]")
            self.assertEqual(lines[1].replace(" ", ""), "seed=1")
            self.assertEqual(lines[2].replace(" ", ""), "task=23")

    def testMapConfigWrite(self):
        """
        Tests the map config output to check output is correct.
        """
        self.coal.add_historical_map(
            fine_file="sample/SA_sample_fine_historical1.tif",
            coarse_file="sample/SA_sample_coarse_historical1.tif",
            time=1,
            rate=0.5,
        )
        self.coal.add_historical_map(
            fine_file="sample/SA_sample_fine_historical2.tif",
            coarse_file="sample/SA_sample_coarse_historical2.tif",
            time=4,
            rate=0.7,
        )
        self.coal.create_map_config("output/mapconf2.txt")
        with open("output/mapconf2.txt", "r") as mapconf:
            lines = mapconf.readlines()
            lines = [x.strip() for x in lines]
            self.assertEqual(lines[21], "[sample_grid]")
            self.assertEqual(
                lines[22].replace(" ", ""),
                "path=null",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[27].replace(" ", ""),
                "[fine_map]",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[28].replace(" ", ""),
                "path=sample/SA_sample_fine.tif",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[29].replace(" ", ""),
                "x=13",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[30].replace(" ", ""),
                "y=13",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[31].replace(" ", ""),
                "x_off=0",
                msg="Config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[32].replace(" ", ""),
                "y_off=0",
                msg="Config file doesn't produce expected output.",
            )

    def testConfigOverwrite(self):
        """Tests that the config file overwrites the output."""
        config_output = os.path.join("output", "conf_test_create.txt")
        config_output2 = os.path.join("output", "notexist", "conf_test_create2.txt")
        sim = Simulation(logging_level=logging.CRITICAL)
        sim.set_simulation_parameters(1, 23, "output", 0.1, 4, 4, 1, 1.0, max_time=200, spatial=False)
        self.assertFalse(os.path.exists(os.path.dirname(config_output2)))
        sim.write_config(config_output2)
        self.assertTrue(os.path.exists(os.path.dirname(config_output2)))
        with open(config_output, "a"):
            pass
        sim = Simulation(logging_level=logging.CRITICAL)
        sim.set_simulation_parameters(
            1,
            23,
            "output",
            0.1,
            4,
            4,
            1,
            1.0,
            max_time=200,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
            restrict_self=True,
        )
        sim.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        sim.sample_map.file_name = None
        sim.coarse_map.file_name = None
        sim.add_sample_time(0.0)
        sim.add_sample_time(1.0)
        sim.set_speciation_rates([0.1, 0.2])
        self.assertTrue(os.path.exists(config_output))
        sim.write_config(config_output)
        self.assertTrue(os.path.exists(config_output))
        with open(config_output, "r") as mapconf:
            lines = mapconf.readlines()
            lines = [x.strip() for x in lines]
            self.assertEqual(lines[0], "[main]")
            self.assertEqual(lines[1].replace(" ", ""), "seed=1")
            self.assertEqual(lines[2].replace(" ", ""), "task=23")

    def testTimeConfigWrite(self):
        """
        Tests the map config writing is correct.
        """
        with open("output/conf1.txt", "r") as f:
            lines = f.readlines()
            lines = [x.strip().replace(" ", "") for x in lines]
            self.assertEqual(
                lines[17],
                "[times]",
                msg="Time config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[18],
                "time0=0.0",
                msg="Time config file doesn't produce expected output.",
            )
            self.assertEqual(
                lines[19],
                "time1=1.0",
                msg="Time config file doesn't produce expected output.",
            )

    def testConfigWrite(self):
        """Tests that the config parser correctly writes all simulation parameters to memory."""
        coal = Simulation(logging_level=logging.CRITICAL)
        coal.set_simulation_parameters(
            seed=1,
            task=23,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=4,
            tau=4,
            deme=1,
            sample_size=1.0,
            max_time=200,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
        )
        coal.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        coal.add_sample_time(0.0)
        coal.add_sample_time(1.0)
        coal.set_speciation_rates([0.1, 0.2])
        coal.write_config("output/conf1b.txt")
        coal.write_config("output/output2/conf1b.txt")
        self.assertTrue(os.path.exists("output/conf1b.txt"))
        self.assertTrue(os.path.exists("output/output2/conf1b.txt"))
        ref_config_parser = ConfigParser()
        ref_config_parser.read("output/conf1b.txt")
        for section in ["sample_grid", "fine_map", "coarse_map", "main"]:
            self.assertTrue(ref_config_parser.has_section(section))

    def testConfigWriteErrors(self):
        """Tests that the config parser correctly writes all simulation parameters to memory."""
        coal = Simulation(logging_level=logging.CRITICAL)
        with self.assertRaises(RuntimeError):
            coal.create_config()
        coal.set_simulation_parameters(
            seed=1,
            task=23,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=4,
            tau=4,
            deme=1,
            sample_size=1.0,
            max_time=200,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="fat-tail",
        )
        with self.assertRaises(RuntimeError):
            coal.create_map_config()
        coal.set_map_files(
            "null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        coal.add_historical_map(
            fine_file="sample/SA_sample_fine_historical1.tif",
            coarse_file="sample/SA_sample_coarse_historical1.tif",
            time=1,
            rate=0.5,
        )
        coal.add_historical_map(
            fine_file="sample/SA_sample_fine_historical2.tif",
            coarse_file="sample/SA_sample_coarse_historical2.tif",
            time=4,
            rate=0.7,
        )
        coal.rates_list = [0.5]
        with self.assertRaises(ValueError):
            coal.write_config("output/conf1c.txt")

    def testConfigRead(self):
        """Tests that the config parser correctly reads simulation parameters from the file."""
        coal = Simulation(logging_level=logging.CRITICAL)
        coal.load_config(os.path.join("sample", "conf_example1.txt"))
        self.assertEqual(1, coal.seed)
        self.assertEqual(23, coal.task)
        self.assertEqual("output", coal.output_directory)
        self.assertEqual(0.1, coal.min_speciation_rate)
        self.assertEqual(4, coal.sigma)
        self.assertEqual(4, coal.tau)
        self.assertEqual(1.0, coal.sample_size)
        self.assertEqual(200, coal.max_time)
        self.assertEqual(1, coal.dispersal_relative_cost)
        self.assertEqual(1, coal.min_num_species)
        self.assertEqual("fat-tail", coal.dispersal_method)
        self.assertEqual("null", coal.sample_map.file_name)
        self.assertEqual("pycoalescence/tests/sample/SA_sample_fine.tif", coal.fine_map.file_name)
        self.assertEqual("pycoalescence/tests/sample/SA_sample_coarse.tif", coal.coarse_map.file_name)
        self.assertEqual(
            "pycoalescence/tests/sample/SA_sample_fine_pristine1.tif",
            coal.historical_fine_list[0],
        )
        self.assertEqual(
            "pycoalescence/tests/sample/SA_sample_fine_pristine2.tif",
            coal.historical_fine_list[1],
        )
        self.assertEqual(
            "pycoalescence/tests/sample/SA_sample_coarse_pristine1.tif",
            coal.historical_coarse_list[0],
        )
        self.assertEqual(
            "pycoalescence/tests/sample/SA_sample_coarse_pristine2.tif",
            coal.historical_coarse_list[1],
        )
        self.assertEqual([10, 10], coal.times_list)
        self.assertEqual([0.5, 0.5], coal.rates_list)
        self.assertEqual([0.0, 1.0], coal.times)
        self.assertEqual([0.1, 0.2], coal.speciation_rates)


class TestSimulationSetMaps(unittest.TestCase):
    """
    Tests that the basic set_map_file() function works as intended and runs a very basic simulation.
    """

    @classmethod
    def setUpClass(cls):
        """
        Creates the coalescence object and runs the setup for the map file
        """
        cls.c = Simulation()
        cls.c.set_map("null", 10, 10)
        cls.c.set_simulation_parameters(
            seed=1,
            task=12,
            output_directory="output",
            min_speciation_rate=0.01,
            sigma=2,
        )
        cls.c.run()

    def testMapFilesSetCorrectly(self):
        """
        Tests that the maps files are set correctly.
        """
        self.assertEqual(self.c.fine_map.file_name, "null")
        self.assertEqual(self.c.fine_map.x_size, 10)
        self.assertEqual(self.c.fine_map.y_size, 10)
        self.assertEqual(self.c.coarse_map.file_name, "none")
        self.assertEqual(self.c.coarse_map.x_size, 10)
        self.assertEqual(self.c.coarse_map.x_size, 10)
        self.assertEqual(self.c.sample_map.file_name, "null")
        self.assertEqual(self.c.sample_map.x_size, 10)
        self.assertEqual(self.c.sample_map.y_size, 10)

    def testSimulationCompletes(self):
        """
        Tests that the simulation completes successfully and outputs as intended.
        """
        self.assertEqual(4, self.c.get_species_richness())

    def testOrdersHistoricalMaps(self):
        """
        Tests that the historical maps are correctly re-ordered.
        """
        historical_maps_fine = ["mapb", "mapa", "mapc", "mapd"]
        historical_maps_coarse = ["mapcb", "mapca", "mapcc", "mapcd"]
        times = [10, 0, 11, 14]
        rates = [0.0, 0.2, 0.9, 0.3]
        s = Simulation()
        s.historical_fine_list = historical_maps_fine
        s.historical_coarse_list = historical_maps_coarse
        s.times_list = times
        s.rates_list = rates
        s.sort_historical_maps()
        expected_fine = ["mapa", "mapb", "mapc", "mapd"]
        expected_coarse = ["mapca", "mapcb", "mapcc", "mapcd"]
        expected_times = [0, 10, 11, 14]
        expected_rates = [0.2, 0.0, 0.9, 0.3]
        self.assertListEqual(expected_fine, s.historical_fine_list)
        self.assertListEqual(expected_coarse, s.historical_coarse_list)
        self.assertListEqual(expected_times, s.times_list)
        self.assertListEqual(expected_rates, s.rates_list)


@unittest.skipIf(sys.version[0] == "2", "Skipping Python 3.x tests")
class TestLoggingOutputsCorrectly(unittest.TestCase):
    """Basic test for expected logging outputs."""

    def testOutputStreamerInfo(self):
        """
        Tests that info output streaming works as intended (skipping the timing information)
        """
        log_stream = StringIO()
        with open("sample/log_12_2.txt", "r") as content_file:
            expected_log = content_file.read().replace("\r", "\n").split("\n")[:-6]
            expected_log[0] = expected_log[0].format(pycoalescence_version)
        s = Simulation(logging_level=logging.INFO, stream=log_stream)
        s.set_simulation_parameters(seed=2, task=12, output_directory="output", min_speciation_rate=0.1)
        s.set_map("null", 10, 10)
        s.run()
        log = log_stream.getvalue().replace("\r", "\n").split("\n")[:-6]
        self.assertEqual(expected_log, log)

    def testOutputStreamerWarning(self):
        """
        Tests that warning output streaming works as intended.
        """
        log_stream = StringIO()
        s = Simulation(logging_level=logging.WARNING, stream=log_stream)
        s.set_simulation_parameters(seed=3, task=12, output_directory="output", min_speciation_rate=0.1)
        s.set_map("null", 10, 10)
        s.finalise_setup()
        s.run_coalescence()
        self.assertEqual("", log_stream.getvalue())

    def testOutputStreamerCritical(self):
        """
        Tests that info output streaming works as intended.
        """
        log_stream = StringIO()
        s = Simulation(logging_level=logging.CRITICAL, stream=log_stream)
        s.set_simulation_parameters(seed=4, task=12, output_directory="output", min_speciation_rate=0.1)
        s.set_map("null", 10, 10)
        s.finalise_setup()
        s.run_coalescence()
        self.assertEqual("", log_stream.getvalue())


@skipLongTest
class TestInitialCountSuccess(unittest.TestCase):
    """
    Tests that the initial count is correct
    """

    def testInitialCountNoCritical(self):
        """
        Tests that the initial count is successful by catching the output of the critical logging.
        """
        log_stream = StringIO()
        s = Simulation(logging_level=logging.CRITICAL, stream=log_stream)
        s.set_simulation_parameters(seed=5, task=12, output_directory="output", min_speciation_rate=0.1)
        s.set_map_files(sample_file="null", fine_file="sample/large_fine.tif")
        s.sample_map.x_size = 10
        s.sample_map.y_size = 10
        s.fine_map.x_offset = 100
        s.fine_map.y_offset = 120
        s.finalise_setup()
        s.run_coalescence()
        self.assertEqual("", log_stream.getvalue())


class TestSimulationDimensionsAndOffsets(unittest.TestCase):
    """Test the dimension detection and offsets of Simulation."""

    @classmethod
    def setUpClass(cls):
        cls.coal = Simulation()
        cls.coal.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        cls.coal2 = Simulation()
        cls.coal2.set_map_files(
            sample_file="null",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )

    def testFineMapDimensions(self):
        """Checks that the dimensions and offsets are properly calculated."""
        self.assertEqual(self.coal.fine_map.x_offset, 0)
        self.assertEqual(self.coal.fine_map.y_offset, 0)
        self.assertAlmostEqual(self.coal.fine_map.x_res, 0.00833308, 5)
        self.assertAlmostEqual(self.coal.fine_map.y_res, -0.00833308, 5)
        self.assertEqual(self.coal.fine_map.x_size, 13)
        self.assertEqual(self.coal.fine_map.y_size, 13)

    def testFineMapDimensionsNull(self):
        """
        Checks that the dimensions and offsets are properly calculated when there is a null map provided as the
        samplemask.
        """
        self.assertEqual(self.coal2.fine_map.x_offset, 0)
        self.assertEqual(self.coal2.fine_map.y_offset, 0)
        self.assertAlmostEqual(self.coal2.fine_map.x_res, 0.00833308, 5)
        self.assertAlmostEqual(self.coal2.fine_map.y_res, -0.00833308, 5)
        self.assertEqual(self.coal2.fine_map.x_size, 13)
        self.assertEqual(self.coal2.fine_map.y_size, 13)

    def testCoarseMapDimensions(self):
        """Checks that the dimensions and offsets are properly calculated."""
        self.assertEqual(self.coal.coarse_map.x_offset, 11)
        self.assertEqual(self.coal.coarse_map.y_offset, 14)
        self.assertAlmostEqual(self.coal.coarse_map.x_res, 0.00833308, 5)
        self.assertAlmostEqual(self.coal.coarse_map.y_res, -0.00833308, 5)
        self.assertEqual(self.coal.coarse_map.x_size, 35)
        self.assertEqual(self.coal.coarse_map.y_size, 41)

    def testCoarseMapDimensionsNull(self):
        """
        Checks that the dimensions and offsets are properly calculated when there is a null map provided as the
        samplemask.
        """
        self.assertEqual(self.coal2.coarse_map.x_offset, 11)
        self.assertEqual(self.coal2.coarse_map.y_offset, 14)
        self.assertAlmostEqual(self.coal2.coarse_map.x_res, 0.00833308, 5)
        self.assertAlmostEqual(self.coal2.coarse_map.y_res, -0.00833308, 5)
        self.assertEqual(self.coal2.coarse_map.x_size, 35)
        self.assertEqual(self.coal2.coarse_map.y_size, 41)

    def testSimStart(self):
        """Checks that the correct exceptions are raised when simulation is started without being properly setup."""
        with self.assertRaises(RuntimeError):
            self.coal.run_coalescence()

    def testDimensionsCheckingError(self):
        """Tests that an error is raised if the dimensions don't match."""
        sim = Simulation()
        sim.fine_map = Map(file=os.path.join("sample", "SA_sample_fine.tif"))
        sim.fine_map.set_dimensions()
        with self.assertRaises(ValueError):
            sim.check_dimensions_match_fine(map_to_check=Map(file=os.path.join("sample", "SA_sample_coarse.tif")))


class TestSimulationExtremeSpeciation(unittest.TestCase):
    """Tests extreme speciation values to ensure that either 1 or maximal numbers of species are produced."""

    def testZeroSpeciation(self):
        """Tests that running a simulation with a zero speciation rate produces a single species."""
        c = Simulation()
        c.set_simulation_parameters(
            seed=1,
            task=17,
            output_directory="output",
            min_speciation_rate=0.0,
            sigma=2.0,
            tau=1,
            deme=1,
            sample_size=1,
            max_time=4,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            landscape_type=False,
        )
        c.set_map("null", 10, 10)
        c.run()
        self.assertEqual(c.get_species_richness(), 1)

    def testMaxSpeciation(self):
        """Tests that running a simulation with a zero speciation rate produces a single species."""
        c = Simulation()
        c.set_simulation_parameters(
            seed=1,
            task=18,
            output_directory="output",
            min_speciation_rate=1.0,
            sigma=2.0,
            tau=1,
            deme=1,
            sample_size=1,
            max_time=4,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            landscape_type=False,
        )
        c.set_map("null", 10, 10)
        c.run()
        self.assertEqual(c.get_species_richness(), 100)

    def testApplySpeciation(self):
        """Tests that speciation rates can be applied post-simulation."""
        c = Simulation()
        c.set_simulation_parameters(
            seed=2,
            task=18,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=2.0,
            tau=1,
            deme=1,
            sample_size=1,
            max_time=4,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            landscape_type=False,
        )
        c.set_map("null", 10, 10)
        c.finalise_setup()
        self.assertTrue(c.run_coalescence())
        c.apply_speciation_rates(speciation_rates=[0.1, 0.5, 0.9999])
        self.assertEqual(25, c.get_species_richness(reference=1))
        self.assertEqual(69, c.get_species_richness(reference=2))
        self.assertEqual(100, c.get_species_richness(reference=3))


class TestSimulationMapDensityReading(unittest.TestCase):
    """
    Tests that the density estimation is relatively accurate and the reading actual density from a map is accurate.
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the coalescence object for referencing the map objects
        """
        cls.c = Simulation()
        cls.c.set_simulation_parameters(
            seed=1,
            task=36,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=64000,
            sample_size=0.00005,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
        )
        cls.c.set_map_files(sample_file="sample/large_mask.tif", fine_file="sample/large_fine.tif")

    def testActualDensity(self):
        """
        Tests the actual density
        """
        self.assertEqual(
            self.c.grid_density_actual(0, 0, self.c.sample_map.x_size, self.c.sample_map.y_size),
            531,
        )

    def testEstimateDensity(self):
        """
        Tests the estimate density for the sample grid
        """
        self.assertEqual(
            self.c.grid_density_estimate(0, 0, self.c.sample_map.x_size, self.c.sample_map.y_size),
            375,
        )

    def testFineAverageMap(self):
        """
        Tests the average density of the fine map is correct.
        """
        self.assertAlmostEqual(self.c.get_average_density(), 38339.499790687034, 3)

    def testCountIndividuals(self):
        """Tests that the count of numbers of individuals is roughly accurate"""
        self.assertTrue(self.c.count_individuals() - 12381 < 2200)
        self.assertTrue(self.c.count_individuals() - 12381 < 2200)
        c = Simulation()
        c.set_simulation_parameters(
            seed=1,
            task=36,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=64000,
            sample_size=0.00005,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
        )
        c.set_map_files(sample_file="null", fine_file="sample/large_fine.tif")
        self.assertAlmostEqual(5907177.600000001, c.count_individuals(), places=2)

    def testSampleMapMatchingTest(self):
        """Tests that sample map equals the sample grid."""
        self.assertEqual(False, self.c.check_sample_map_equals_sample_grid())

    def testImportingMapArrays(self):
        """Tests that importing the map arrays works correctly."""
        self.c.import_fine_map_array()
        self.assertAlmostEqual(1845993, np.sum(self.c.fine_map_array), places=2)

    def testImportingNullMapArrays(self):
        """Tests that importing the map arrays works correctly."""
        sim = Simulation()
        sim.set_map("null", 10, 10)
        sim.import_fine_map_array()
        sim.import_sample_map_array()
        self.assertEqual(100, np.sum(sim.fine_map_array))
        self.assertEqual(100, np.sum(sim.sample_map_array))
        self.assertEqual(100, sim.grid_density_estimate(0, 0, 2, 2))
        self.assertEqual(4, sim.grid_density_actual(0, 0, 2, 2))
        sim = Simulation()
        sim.uses_spatial_sampling = True
        sim.set_map_files(
            sample_file=os.path.join("sample", "SA_sample_fine.tif"),
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
        )
        sim.import_sample_map_array()
        self.assertEqual(38098, np.sum(sim.sample_map_array))


class TestHistoricalMapsAlterResult(unittest.TestCase):
    """
    Makes sure that historical maps correctly alter the result of the simulation.
    """

    @classmethod
    def setUpClass(cls):
        cls.base_sim = Simulation()
        cls.hist_sim = Simulation()
        cls.base_sim.set_simulation_parameters(
            seed=4,
            task=17,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=2,
            sample_size=0.1,
        )
        cls.base_sim.set_map("sample/SA_sample_fine.tif")
        cls.base_sim.run()
        cls.hist_sim.set_simulation_parameters(
            seed=4,
            task=18,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=2,
            sample_size=0.1,
        )
        cls.hist_sim.set_map("sample/SA_sample_fine.tif")
        cls.hist_sim.add_historical_map(
            fine_file="sample/example_historical_fine.tif",
            coarse_file="none",
            time=10,
            rate=0.2,
        )
        cls.hist_sim.run()
        cls.hist_sim2 = Simulation()
        cls.hist_sim2.set_simulation_parameters(
            seed=4,
            task=19,
            output_directory="output",
            min_speciation_rate=0.1,
            sigma=2,
            sample_size=0.1,
        )
        cls.hist_sim2.set_map("sample/SA_sample_fine.tif")
        cls.hist_sim2.add_historical_map(
            fine_file="sample/example_historical_fine.tif",
            coarse_file="none",
            time=10,
            rate=0.2,
        )
        cls.hist_sim2.add_historical_map(fine_file="sample/SA_sample_fine.tif", coarse_file="none", time=20, rate=0.2)
        cls.hist_sim2.run()

    def testSpeciesRichnessDiffer(self):
        """
        Tests that the species richness differs between the two simulations
        """
        self.assertNotEqual(self.base_sim.get_species_richness(), self.hist_sim.get_species_richness())
        self.assertNotEqual(self.hist_sim.get_species_richness(), self.hist_sim2.get_species_richness())
        self.assertEqual(2673, self.base_sim.get_species_richness())
        self.assertEqual(2515, self.hist_sim2.get_species_richness())
        self.assertEqual(2450, self.hist_sim.get_species_richness())


@skipLongTest
class TestExpansionOverTime(unittest.TestCase):
    """Tests that large expansions over time are dealt with properly when sampling multiple time points."""

    @classmethod
    def setUpClass(cls):
        """Run the simulation for expansion over time."""
        cls.sim = Simulation(logging_level=60)
        cls.sim.set_simulation_parameters(
            seed=5,
            task=17,
            output_directory="output",
            min_speciation_rate=0.0001,
            sigma=1,
            deme=100,
            sample_size=1.0,
            landscape_type="infinite",
        )
        cls.sim.set_map_files("null", "sample/null.tif", "sample/null_large.tif")
        cls.sim.add_historical_map("sample/null.tif", "sample/null_large.tif", time=500, rate=0.5)
        cls.sim.add_sample_time([0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])
        cls.sim.run()

    def testSpeciesRichnessAtTimes(self):
        """Checks the species richness is correct for each time point."""
        self.assertEqual(336, self.sim.get_species_richness(1))
        self.assertEqual(332, self.sim.get_species_richness(2))
        self.assertEqual(332, self.sim.get_species_richness(3))
        self.assertEqual(357, self.sim.get_species_richness(4))
        self.assertEqual(340, self.sim.get_species_richness(5))
        self.assertEqual(343, self.sim.get_species_richness(6))


class TestSimulationParameters(unittest.TestCase):
    """Tests that parameters are correctly set."""

    def testAddingTimes(self):
        """Tests times are correctly added."""
        sim = Simulation()
        sim.set_simulation_parameters(
            seed=10000,
            task=1000000,
            output_directory="output",
            min_speciation_rate=0.1,
            times=0.1,
        )
        self.assertEqual([0.0, 0.1], sim.times)
        sim = Simulation()
        sim.set_simulation_parameters(
            seed=10000,
            task=1000000,
            output_directory="output",
            min_speciation_rate=0.1,
            times=[10, 100],
        )
        self.assertEqual([0.0, 10.0, 100.0], sim.times)
        sim = Simulation()
        sim.set_simulation_parameters(
            seed=10000,
            task=1000000,
            output_directory="output",
            min_speciation_rate=0.1,
            times=[100.0],
        )
        self.assertEqual([0.0, 100.0], sim.times)

    def testProtractedParameters(self):
        """Tests that the protracted paramaters are obtained correctly."""

        sim = Simulation()
        sim.protracted = True
        self.assertTrue(sim.get_protracted())
        sim = Simulation()
        self.assertEqual(27, sim.run_simple(6, 17, "output", 0.1, 4, 10))
        self.assertFalse(sim.get_protracted())

    def raise_ioerror(self, path):
        """For mocking modules."""
        raise IOError

    def return_true(self):
        """For mocking modules"""
        return True

    @patch.object(CoalescenceTree, "set_database", raise_ioerror)
    @patch.object(CoalescenceTree, "is_protracted", return_true)
    def testGetProtractedFailure(self):
        """Tests that errors while opening the database are handled correctly."""
        sim = Simulation()
        sim.output_database = os.path.join("sample", "data_old1.db")
        self.assertTrue(sim.get_protracted())

    def testCheckReproductionMap(self):
        """Tests that the reproduction map configuration is correctly identified"""
        sim = Simulation()
        sim.reproduction_map.file_name = "path"
        sim.coarse_map.file_name = "path"
        with self.assertRaises(ValueError):
            sim.check_reproduction_map()

    def testCheckDeathMap(self):
        """Tests that the death map configuration is correctly identified"""
        sim = Simulation()
        sim.death_map.file_name = "path"
        sim.coarse_map.file_name = "path"
        with self.assertRaises(ValueError):
            sim.check_death_map()

    def mock_check_map(self):
        """For mocking checking if map exists"""
        pass

    @patch.object(Simulation, "check_dispersal_map", mock_check_map)
    def testAddDispersalMap(self):
        """Tests that a dispersal map can be added correctly."""
        sim = Simulation()
        sim.add_dispersal_map("a_map.tif")
        self.assertEqual("a_map.tif", sim.dispersal_map.file_name)
        m = Map()
        m.file_name = "a_map.tif"
        sim = Simulation()
        sim.add_dispersal_map(m)
        self.assertEqual("a_map.tif", sim.dispersal_map.file_name)

    @patch.object(Simulation, "check_reproduction_map", mock_check_map)
    def testAddReproductionMap(self):
        """Tests that a dispersal map can be added correctly."""
        sim = Simulation()
        sim.add_reproduction_map("a_map.tif")
        self.assertEqual("a_map.tif", sim.reproduction_map.file_name)
        m = Map()
        m.file_name = "a_map.tif"
        sim = Simulation()
        sim.add_reproduction_map(m)
        self.assertEqual("a_map.tif", sim.reproduction_map.file_name)

    @patch.object(Simulation, "check_death_map", mock_check_map)
    def testAddDeathMap(self):
        """Tests that a death map can be added correctly."""
        sim = Simulation()
        sim.add_death_map("a_map.tif")
        self.assertEqual("a_map.tif", sim.death_map.file_name)
        m = Map()
        m.file_name = "a_map.tif"
        sim = Simulation()
        sim.add_death_map(m)
        self.assertEqual("a_map.tif", sim.death_map.file_name)

    def testSpatialErrors(self):
        """Tests that an error is raised for spatial sampling a non-spatial simulation."""
        sim = Simulation()
        with self.assertRaises(ValueError):
            sim.set_simulation_parameters(
                seed=10000,
                task=1000000,
                output_directory="output",
                min_speciation_rate=0.1,
                spatial=False,
                uses_spatial_sampling=True,
            )
        sim = Simulation()
        with self.assertRaises(ValueError):
            sim.set_simulation_parameters(
                seed=10000,
                task=1000000,
                output_directory="output",
                min_speciation_rate=0.1,
                landscape_type="bleh",
            )
        sim = Simulation(logging_level=40)
        sim.set_simulation_parameters(
            seed=10000,
            task=1000000,
            output_directory="output",
            min_speciation_rate=0.1,
        )
        sim.set_simulation_parameters(
            seed=10000,
            task=1000000,
            output_directory="output",
            min_speciation_rate=0.1,
        )

    def testParametersNotSetError(self):
        """Tests that an error is raised when the parameters aren't set."""
        sim = Simulation()
        with self.assertRaises(RuntimeError):
            sim.check_simulation_parameters()
        sim.full_config_file = os.path.join("output", "tmp_output2", "full_config.txt")
        sim.is_setup_param = True
        with self.assertRaises(RuntimeError):
            sim.check_simulation_parameters()
        sim.seed = 1
        # sim.output_directory = None
        with self.assertRaises(RuntimeError):
            sim.check_simulation_parameters()
        sim.output_directory = os.path.join("output", "tmp_output")

        sim.set_map("null", 10, 10)
        sim.check_simulation_parameters()
        self.assertTrue(os.path.exists(sim.output_directory))
        self.assertTrue(os.path.exists(os.path.join("output", "tmp_output2")))

    def testResumeError(self):
        """Tests that resuming fails if the output directory doesn't exist"""
        sim = Simulation()
        with self.assertRaises(IOError):
            sim.resume_coalescence("notadir", 1, 1, 10)

    def testDatabaseSettingErrors(self):
        """Tests that the database can be set correctly."""
        sim = Simulation()
        sim.output_database = os.path.join("sample", "sample.db")
        with self.assertRaises(IOError):
            sim.check_sql_database()
        sim.output_database = os.path.join("sample", "not_a_file.db")
        with self.assertRaises(IOError):
            sim.check_sql_database(expected=True)

    def testRunChecksErrors(self):
        """Tests errors are raised when checks are not complete."""
        sim = Simulation()
        with self.assertRaises(RuntimeError):
            sim.check_file_parameters()

    def testRunCoalescenceCreatesOutput(self):
        """Tests that running coalescence creates output before running."""
        sim = Simulation()
        sim.is_setup_complete = True
        sim.output_directory = os.path.join("output", "tmp_output3")

        class tmp:
            pass

        sim.c_simulation = tmp
        sim.c_simulation.run = self.return_true
        self.assertTrue(sim.run_coalescence())
        self.assertTrue(os.path.exists(os.path.join("output", "tmp_output3")))


class TestRamEstimation(unittest.TestCase):
    """Tests that the RAM estimation works as intended"""

    def testPersistentRamEstimation(self):
        """Tests the calculations for persistent RAM usage."""
        sim = Simulation()
        sim.set_simulation_parameters(1, 1000, output_directory="output", min_speciation_rate=0.1)
        sim.set_map_files(
            "null",
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )
        sim.add_historical_map(
            fine_file=os.path.join("sample", "SA_sample_fine_pristine1.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse_pristine1.tif"),
            time=10,
            rate=0.0,
        )
        self.assertEqual(9689724, sim.persistent_ram_usage())
        sim = Simulation()
        sim.set_simulation_parameters(1, 1000, output_directory="output", min_speciation_rate=0.1)
        sim.set_map_files(
            sample_file=os.path.join("sample", "SA_sample_coarse.tif"),
            fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )
        sim.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine.tif"))
        sim.add_death_map(death_map=os.path.join("sample", "SA_sample_coarse_pristine.tif"))
        sim.add_reproduction_map(reproduction_map=os.path.join("sample", "SA_sample_coarse_pristine.tif"))
        self.assertEqual(30135, sim.persistent_ram_usage())


class TestRamOptimistation(unittest.TestCase):
    """Tests that the RAM optimisation works as intended."""

    def testBasicOptimisation(self):
        """Test that a basic optimisation process works."""
        sim = Simulation()
        sim.set_simulation_parameters(seed=1, task=2, output_directory="output", min_speciation_rate=0.1)
        sim.set_map_files(
            "null",
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )
        sim.optimise_ram(0.0005)
        self.assertEqual(1, sim.grid.x_size)
        expected_dict = {
            "grid_x_size": 1,
            "grid_y_size": 1,
            "sample_x_offset": 0,
            "sample_y_offset": 0,
            "grid_file_name": "set",
        }
        self.assertEqual(expected_dict, sim.get_optimised_solution())
        expected_dict2 = {
            "grid_x_size": 10,
            "grid_y_size": 10,
            "sample_x_offset": 0,
            "sample_y_offset": 0,
            "grid_file_name": "set",
        }
        sim.set_optimised_solution(expected_dict2)
        self.assertEqual(expected_dict2, sim.get_optimised_solution())

    def testOptimisationRaisesError(self):
        """Tests that errors are raised correctly."""
        sim = Simulation()
        sim.set_simulation_parameters(seed=1, task=2, output_directory="output", min_speciation_rate=0.1)
        sim.set_map_files(
            "null",
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )
        sim.deme = 0
        with self.assertRaises(ValueError):
            sim.optimise_ram(0.1)
        sim.deme = 1
        with self.assertRaises(MemoryError):
            sim.optimise_ram(0.00001)
        with self.assertRaises(MemoryError):
            sim.optimise_ram(0.0000001)
        sim = Simulation()
        sim.set_simulation_parameters(seed=1, task=2, output_directory="output", min_speciation_rate=0.1)
        sim.set_map_files(
            "null",
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )


@skipLongTest
class TestSimulationUsingGillespieEquality(unittest.TestCase):
    """
    Tests simulations using the gillespie algorithm match equivalent simulations not using the Gillespie algorithm.
    """

    @classmethod
    def setUpClass(cls):
        cls.baseline_richness_values = []
        speciation_rates = [0.001, 0.01, 0.1, 0.9]
        for seed in range(10, 20):
            baseline_simulation = Simulation(logging_level=50)
            baseline_simulation.set_simulation_parameters(
                seed=seed,
                task=2,
                output_directory="output",
                min_speciation_rate=0.001,
                deme=1,
                sample_size=0.01,
            )
            baseline_simulation.set_map_files(
                sample_file="null",
                fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
                coarse_file="none",
            )
            baseline_simulation.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine2.tif"))
            baseline_simulation.set_speciation_rates(speciation_rates=speciation_rates)
            baseline_simulation.run()
            for ref in range(1, len(speciation_rates) + 1):
                cls.baseline_richness_values.append((ref, baseline_simulation.get_species_richness(ref)))
        cls.gillespie_richness_values = []
        for seed, gillespie_generation in zip(range(10, 20), range(10, 2020, 200)):
            gillespie_simulation = Simulation(logging_level=50)
            gillespie_simulation.set_simulation_parameters(
                seed=seed,
                task=3,
                output_directory="output",
                min_speciation_rate=0.001,
                deme=1,
                sample_size=0.01,
            )
            gillespie_simulation.set_speciation_rates(speciation_rates=speciation_rates)
            gillespie_simulation.set_map_files(
                sample_file="null",
                fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
                coarse_file="none",
            )
            gillespie_simulation.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine2.tif"))
            gillespie_simulation.add_gillespie(gillespie_generation)
            gillespie_simulation.run()
            for ref in range(1, len(speciation_rates) + 1):
                cls.gillespie_richness_values.append((ref, gillespie_simulation.get_species_richness(ref)))

    @staticmethod
    def setupGillespie(**kwargs):
        """Sets up a simple Gillespie simulation."""
        s = Simulation()
        s.set_simulation_parameters(
            seed=12, task=3, output_directory="output", min_speciation_rate=0.001, deme=1, sample_size=0.01, **kwargs
        )
        return s

    def testChecksForGillespieCompability(self):
        """Checks that simulations which can't use Gillespie throw the correct errors."""
        coarse_maps = ["null", os.path.join("sample", "SA_sample_coarse.tif")]
        for coarse_map in coarse_maps:
            s = self.setupGillespie()
            with self.assertRaises(RuntimeError):
                s.add_gillespie(10)
            s.set_map_files(
                sample_file="null",
                fine_file=os.path.join("sample", "SA_sample_fine.tif"),
                coarse_file=coarse_map,
            )
            s.add_dispersal_map("null")
            with self.assertRaises(ValueError):
                s.add_gillespie(10)
        s = self.setupGillespie()
        s.set_map("null", 10, 10)
        s.add_dispersal_map("null")
        with self.assertRaises(ValueError):
            s.add_gillespie(-10)

    def testCheckCanUseGillespie(self):
        """Checks that the Gillespie checks are accurate."""
        s = Simulation()
        with self.assertRaises(RuntimeError):
            s.check_can_use_gillespie()
        s = self.setupGillespie()
        with self.assertRaises(RuntimeError):
            s.check_can_use_gillespie()
        s.set_map("null", 10, 10)
        self.assertFalse(s.check_can_use_gillespie())
        s.add_dispersal_map("null")
        self.assertTrue(s.check_can_use_gillespie())
        s = Simulation()
        s.set_simulation_parameters(
            seed=10,
            task=3,
            output_directory="output",
            min_speciation_rate=0.000001,
            deme=1000,
            sample_size=0.00001,
            spatial=False,
        )
        self.assertFalse(s.check_can_use_gillespie())
        s = Simulation()
        s.set_simulation_parameters(
            seed=10,
            task=3,
            output_directory="output",
            min_speciation_rate=0.000001,
            deme=1000,
            sample_size=0.00001,
            protracted=True,
            min_speciation_gen=10.0,
            max_speciation_gen=100.0,
        )
        s.set_map("null", 10, 10)
        s.add_dispersal_map("null")
        self.assertFalse(s.check_can_use_gillespie())
        s = self.setupGillespie()
        s.set_map_files(
            "null",
            fine_file=os.path.join("sample", "SA_sample_fine.tif"),
            coarse_file=os.path.join("sample", "SA_sample_coarse.tif"),
        )
        s.add_dispersal_map("null")
        self.assertFalse(s.check_can_use_gillespie())

    def testSpeciesRichnessValuesSimilar(self):
        """Checks that the species richness values are similar between implementations of Gillespie."""
        baseline_mean_values = {}
        baseline_all_values = {}
        for i in set(x for x, _ in self.baseline_richness_values):
            vals = [richness for ref, richness in self.baseline_richness_values if ref == i]
            baseline_all_values[i] = vals
            baseline_mean_values[i] = sum(vals) / len(vals)

        gillespie_mean_values = {}
        gillespie_all_values = {}
        for i in set(x for x, _ in self.gillespie_richness_values):
            vals = [richness for ref, richness in self.gillespie_richness_values if ref == i]
            gillespie_all_values[i] = vals
            gillespie_mean_values[i] = sum(vals) / len(vals)
        for k, v in baseline_mean_values.items():
            self.assertAlmostEqual(v, gillespie_mean_values[k], delta=v / 10)


@skipLongTest
class TestSimulationUsingGillespieDeathMaps(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.baseline_richness_values = []
        speciation_rates = [0.001, 0.01, 0.1, 0.9]
        for seed in range(40, 50):
            baseline_simulation = Simulation(logging_level=50)
            baseline_simulation.set_simulation_parameters(
                seed=seed,
                task=2,
                output_directory="output",
                min_speciation_rate=0.001,
                deme=2,
                sample_size=0.01,
            )
            baseline_simulation.set_map_files(
                sample_file="null",
                fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
                coarse_file="none",
            )
            baseline_simulation.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine2.tif"))
            baseline_simulation.add_death_map(os.path.join("sample", "SA_death.tif"))
            baseline_simulation.set_speciation_rates(speciation_rates=speciation_rates)
            baseline_simulation.run()
            for ref in range(1, len(speciation_rates) + 1):
                cls.baseline_richness_values.append((ref, baseline_simulation.get_species_richness(ref)))
        cls.gillespie_richness_values = []
        for seed, gillespie_generation in zip(range(40, 50), range(10, 2020, 200)):
            gillespie_simulation = Simulation(logging_level=50)
            gillespie_simulation.set_simulation_parameters(
                seed=seed,
                task=3,
                output_directory="output",
                min_speciation_rate=0.001,
                deme=2,
                sample_size=0.01,
            )
            gillespie_simulation.set_speciation_rates(speciation_rates=speciation_rates)
            gillespie_simulation.set_map_files(
                sample_file="null",
                fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
                coarse_file="none",
            )
            gillespie_simulation.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine2.tif"))
            gillespie_simulation.add_death_map(os.path.join("sample", "SA_death.tif"))
            gillespie_simulation.add_gillespie(gillespie_generation)
            gillespie_simulation.run()
            for ref in range(1, len(speciation_rates) + 1):
                cls.gillespie_richness_values.append((ref, gillespie_simulation.get_species_richness(ref)))

    def testSpeciesRichnessValuesSimilar(self):
        """Checks that the species richness values are similar between implementations of Gillespie."""
        baseline_mean_values = {}
        baseline_all_values = {}
        for i in set(x for x, _ in self.baseline_richness_values):
            vals = [richness for ref, richness in self.baseline_richness_values if ref == i]
            baseline_all_values[i] = vals
            baseline_mean_values[i] = sum(vals) / len(vals)

        gillespie_mean_values = {}
        gillespie_all_values = {}
        for i in set(x for x, _ in self.gillespie_richness_values):
            vals = [richness for ref, richness in self.gillespie_richness_values if ref == i]
            gillespie_all_values[i] = vals
            gillespie_mean_values[i] = sum(vals) / len(vals)
        for k, v in baseline_mean_values.items():
            self.assertAlmostEqual(v, gillespie_mean_values[k], delta=v / 10)


@skipLongTest
class TestSimulationUsingGillespieLarge(unittest.TestCase):
    """Tests simulations using the gillespie algorithm."""

    @classmethod
    def setUpClass(cls):
        cls.gillespie_simulation = Simulation(logging_level=50)
        cls.gillespie_simulation.set_simulation_parameters(
            seed=21,
            task=3,
            output_directory="output",
            min_speciation_rate=0.000001,
            deme=100,
            sample_size=0.001,
        )
        cls.gillespie_simulation.set_map_files(
            sample_file="null",
            fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
            coarse_file="none",
        )
        cls.gillespie_simulation.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine3.tif"))
        cls.gillespie_simulation.add_gillespie(10)
        cls.gillespie_simulation.run()
        cls.gillespie_simulation2 = Simulation()
        cls.gillespie_simulation2.set_simulation_parameters(
            seed=21,
            task=4,
            output_directory="output",
            min_speciation_rate=0.000001,
            deme=100,
            sample_size=0.001,
        )
        cls.gillespie_simulation2.set_map_files(
            sample_file="null",
            fine_file=os.path.join("sample", "SA_sample_coarse.tif"),
            coarse_file="none",
        )
        cls.gillespie_simulation2.add_dispersal_map(dispersal_map=os.path.join("sample", "dispersal_fine3.tif"))

        cls.gillespie_simulation2.add_gillespie(0)
        cls.gillespie_simulation2.run()

    def testSpeciesRichnessValuesSimilar(self):
        """Checks that the species richness values are similar between implementations of Gillespie."""
        self.assertEqual(208, self.gillespie_simulation.get_species_richness())
        self.assertEqual(208, self.gillespie_simulation2.get_species_richness())



"""
Safely open, close and fetch data from an sqlite connection.

:class:`.SQLiteConnection` contains context management for opening sql connections, plus basic functionality for
detecting existence and structure of databases.
"""
import sqlite3


class SQLiteConnection:
    """
    Class containing context management for opening sqlite3 connections. The file name provided can either be a string
    containing the path to the file, or an sqlite3.Connection object, which will NOT be closed on destruction. This
    provides two points of entry to the system with the same interface.
    """

    def __init__(self, filename):
        """
        Initialises the SQLiteConnection object.

        :param str/sqlite3.Connection filename: Should be either a string containing a path to a file to open,
                                                or an already-open sqlite3.Connection object. If the latter,
                                                the connection will not be closed on destruction. filename can also
                                                be a sqlite3.Cursor object, in which case it will simply be returned.
        """
        self.filename = filename
        self.opened_here = False
        self.database = None

    def __enter__(self):
        """
        Opens the connection to the file and returns an active cursor.

        :return: Active cursor pointing to the file
        :rtype: sqlite3.Cursor
        """
        if isinstance(self.filename, sqlite3.Connection):
            self.database = self.filename
        elif isinstance(self.filename, sqlite3.Cursor):
            return self.filename
        else:
            self.opened_here = True
            try:
                self.database = sqlite3.connect(self.filename)
            except sqlite3.Error as soe:  # pragma: no cover
                self.opened_here = False
                raise soe
        return self.database.cursor()

    def __exit__(self, *args):
        if self.opened_here:
            self.database.commit()
            self.database.close()
            self.database = None


def get_table_names(database):
    """
    Gets a list of all table names in the database.

    :param str/sqlite3.Connection database: the path to the database connection or an already-open database object

    :return: a list of all table names from the database
    :rtype: list
    """
    with SQLiteConnection(database) as cursor:
        return [x[0] for x in cursor.execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()]


def check_sql_table_exist(database, table_name):
    """
    Checks that the supplied table exists in the supplied database.

    :param str/sqlite3.Connection database: the database to check existence in
    :param str table_name: the table name to check for

    :return: true if the table exists
    :rtype: bool
    """
    table_names = [x.lower() for x in get_table_names(database)]
    return table_name.lower() in table_names


def check_sql_column_exists(database, table_name, column_name):
    """
    Checks if the column exists in the database.

    :param str/sqlite3.Connection database: the database to check existence in
    :param str table_name: the table name to check within
    :param str column_name: the column name to check for

    :return: true if the column exists.
    :rtype: bool
    """
    with SQLiteConnection(database) as cursor:
        c = [i[1] for i in cursor.execute("PRAGMA table_info({})".format(table_name)) if i[1] == column_name]
        return len(c) != 0


def fetch_table_from_sql(database, table_name, column_names=False):
    """
    Returns a list of the data contained by the provided table in the database.

    :raises sqlite3.Error: if the table is not contained in the database (protects SQL injections).

    :param str/sqlite3.Connection database: the database to obtain from
    :param str table_name: the table name to fetch data from
    :param bool column_names: if true, return the column names as the first row in the output

    :return: a list of lists, containing all data within the provided table in the database
    """
    try:
        with SQLiteConnection(database) as c:
            if not check_sql_table_exist(c, table_name):
                raise sqlite3.Error("Table {} does not exist in database.".format(table_name))
            c.execute("SELECT * FROM {}".format(table_name))
            output = []
            if column_names:
                output.append([x[0] for x in c.description])
            output.extend([list(x) for x in c.fetchall()])
            return output
    except sqlite3.Error as soe:  # pragma: no cover
        raise IOError("Cannot fetch {} from database {}: {}".format(table_name, database, soe))


def sql_get_max_from_column(database, table_name, column_name):
    """
    Returns the maximum value from the specified column.

    :param str/sqlite3.Connection database: the database to fetch from
    :param str table_name: the table name to attain
    :param str column_name: the column name to obtain from
    :return:
    """
    with SQLiteConnection(database) as cursor:
        if not check_sql_table_exist(cursor, table_name):
            raise sqlite3.Error("Table {} does not exist in database.".format(table_name))
        if not check_sql_column_exists(cursor, table_name, column_name):
            raise sqlite3.Error("Column {} does not exist in {}.".format(column_name, table_name))
        ex_str = "SELECT max({}) FROM {}".format(column_name, table_name)
        cursor.execute(ex_str)
        return cursor.fetchone()[0]



"""
pycoalescence provides the facilities for running spatially explicit neutral coalescence ecological simulations
and performing basic analysis of the simulation outputs. The program requires necsim to function properly.

"""
__version__ = "1.2.9.post2"

import logging

__author__ = "Samuel Thompson"
__copyright__ = "Copyright 2016, pycoalescence"
__credits__ = ["Samuel Thompson"]
__license__ = "MIT"
__maintainer__ = "Samuel Thompson"
__email__ = "thompsonsed@gmail.com"

try:
    from pycoalescence.simulation import Simulation, Map
    from pycoalescence.coalescence_tree import CoalescenceTree
    from pycoalescence.system_operations import set_logging_method
    from pycoalescence.merger import Merger
    from pycoalescence.dispersal_simulation import DispersalSimulation
    from pycoalescence.landscape_metrics import LandscapeMetrics

    # Support for no scipy installed
    try:
        from pycoalescence.fragments import FragmentedLandscape
    except ImportError as ie:  # pragma: no cover
        logging.warning("Cannot generate fragmented landscapes: {}".format(ie))
except ImportError as ie:  # pragma: no cover
    logging.warning(ie)
    logging.warning("Cannot import package - are you sure that it has been installed correctly?")



"""
Run spatially explicit neutral simulations on provided landscapes with support for a wide range of scenarios and
parameters. Detailed :ref:`here <performing_simulations>`.

The main class is :class:`.Simulation`, which contains routines for setting up and running simulations, plus basic tree
generation after simulations have been completed.

:input:
    - Simulation parameters (such as dispersal kernel, speciation rate)
    - Map files representing density over space
    - [optional] map files representing relative reproductive ability
    - [optional] map files representing dispersal potential
    - [optional] historical density map files

:output:
    - Database containing generated coalescence tree, simulation parameters and basic biodiversity metrics.
    - If the simulation does not complete, will instead dump data to a Dump_main_*_*.csv file for resuming simulations.
"""
from __future__ import absolute_import
from __future__ import print_function

import copy
import logging
import os
import types

import numpy as np

global sqlite_import

# noinspection PyCompatibility
from configparser import ConfigParser, NoOptionError

try:
    from io import StringIO
# Python 2.x support
except ImportError as ie:  # pragma: no cover
    from cStringIO import StringIO

try:
    NumberTypes = (types.IntType, types.LongType, types.FloatType, types.ComplexType)
except AttributeError:
    # No support for complex numbers compile
    NumberTypes = (int, float)

try:
    from osgeo import gdal
except ImportError as ie:  # pragma: no cover
    try:
        import gdal
    except ImportError:
        gdal = None
        logging.warning("Problem importing gdal modules: {}".format(ie))

import sqlite3

from pycoalescence.coalescence_tree import CoalescenceTree

from pycoalescence.necsim import CNSESimulation, CSpatialSimulation, CPNSESimulation, CPSpatialSimulation
from pycoalescence.future_except import FileNotFoundError, FileExistsError
from pycoalescence.landscape import Landscape
from pycoalescence.map import Map
from pycoalescence.system_operations import check_parent


class Simulation(Landscape):
    """
    A class containing routines to set up and run simulations, including detecting map dimensions from tif files.
    """

    def __init__(self, logging_level=logging.WARNING, log_output=None, **kwargs):
        """
        Sets up the simulation object, setting the logging object and assigning default variables.

        :param int logging_level: the minimal level of logging to display
        :param string log_output: a file to output log data to. Use None for writing to console.
        :param kwargs: additional keyword arguments to supply to logging.basicConfig()
        """
        # All the parameters used for the simulation
        Landscape.__init__(self)
        # Contains the C object for performing simulations
        # This will be set depending on the simulation type
        self.c_simulation = None
        self._fine_map_sum_res = None
        self.count_total = None
        # Create the logging object and pass on the required arguments for its creation
        passedKwards = {k: v for k, v in kwargs.items() if "stream" not in k}
        self.logger = logging.Logger("necsim", **passedKwards)
        self._create_logger(file=log_output, logging_level=logging_level, **kwargs)
        self.output_directory = ""
        self.pause_directory = ""
        self.config = ConfigParser()
        self.config_string = StringIO()
        self.full_config_file = None
        self.is_setup = False
        self.is_setup_param = False
        self.is_setup_complete = False
        self.seed = 0
        self.min_speciation_rate = 0
        self.sigma = 2
        self.deme = 1
        self.sample_size = 1
        self.max_time = 100
        self.dispersal_relative_cost = 1  # the relative cost of moving through non-matrix
        self.task = 0
        self.min_num_species = 1
        self.tau = 0
        self.times = None
        self.speciation_rates = []
        # parameters for the application of speciation rates
        self.is_setup_speciation = False
        self.speciation_file = ""
        self.record_spatial = False
        self.record_fragments = False
        # The output database location
        self.output_database = None
        self.dispersal_method = None
        self.m_prob = 0
        self.cutoff = 0
        self.restrict_self = False
        # The protracted variables
        self.protracted = False
        self.min_speciation_gen = 0.0
        self.max_speciation_gen = 0.0
        # This variable stores whether a new config file has been opened or not
        self.config_open = False
        # Grid dimensions sizing
        self.grid = Map()
        # Dispersal map data
        self.dispersal_map = Map()
        self.death_map = Map()
        self.reproduction_map = Map()
        self.fine_map_array = None
        self.sample_map_array = None
        self.fine_map_average = None
        self.is_spatial = True
        self.uses_spatial_sampling = False
        self.uses_gillespie = False
        self.gillespie_threshold = None

    def __del__(self):
        """Safely destroys the logger and the C++ objects."""
        handlers = copy.copy(self.logger.handlers)
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)
        self.logger = None
        self.c_simulation = None

    def add_sample_time(self, time):
        """
        Adds an extra sample time to the list of times.

        This allows for multiple temporal sample points from within the same simulation.

        :param time: the sample time to add
        """
        if isinstance(time, list):
            [self.add_sample_time(x) for x in time]
        else:
            if not self.times or len(self.times) == 0:
                self.times = [0.0]
            self.times.append(time)

    def set_config_file(self, output_file=None):
        """
        Sets the config file to the output, over-writing any existing config file that has been stored.

        :param output_file: path to config file to output to
        """
        if self.full_config_file and output_file != self.full_config_file and output_file:
            self.logger.warning(
                "Config file has already been set to {}, changing to {}.".format(self.full_config_file, output_file)
            )
        if output_file is not None:
            self.full_config_file = output_file

    def load_config(self, config_file):
        """
        Loads the config file by reading the lines in order.

        :param str config_file: the config file to read in.
        """
        # New method using ConfigParser
        self.config.read(config_file)
        self.full_config_file = config_file
        self.seed = self.config.getint("main", "seed")
        try:
            self.task = self.config.getint("main", "task")
        except NoOptionError:
            self.task = self.config.getint("main", "job_type")
        self.output_directory = self.config.get("main", "output_directory")
        self.min_speciation_rate = self.config.getfloat("main", "min_spec_rate")
        self.sigma = self.config.getfloat("main", "sigma", fallback=1.0)
        self.tau = self.config.getfloat("main", "tau", fallback=1.0)
        self.dispersal_relative_cost = self.config.getint("main", "dispersal_relative_cost", fallback=1.0)
        self.deme = self.config.getint("main", "deme", fallback=1.0)
        self.sample_size = self.config.getfloat("main", "sample_size", fallback=1.0)
        self.max_time = self.config.getint("main", "max_time", fallback=1.0)
        self.min_num_species = self.config.getint("main", "min_species")
        if self.config.has_section("protracted"):
            self.min_speciation_gen = self.config.getfloat("protracted", "min_speciation_gen")
            self.max_speciation_gen = self.config.getfloat("protracted", "max_speciation_gen")
            self.protracted = True
        if self.config.has_section("spec_rates"):
            opts = self.config.options("spec_rates")
            for each in opts:
                self.speciation_rates.append(self.config.getfloat("spec_rates", each))
        if self.config.has_section("times"):
            opts = self.config.options("times")
            self.times = []
            for each in opts:
                self.times.append(self.config.getfloat("times", each))
        if self.config.has_section("sample_grid"):
            self.sample_map.set_dimensions(
                self.config.get("sample_grid", "path"),
                self.config.getint("sample_grid", "x", fallback=0),
                self.config.getint("sample_grid", "y", fallback=0),
                0,
                0,
            )
            self.uses_spatial_sampling = self.config.getboolean("sample_grid", "uses_spatial_sampling", fallback=False)
            self.fine_map.set_dimensions(
                self.config.get("fine_map", "path"),
                self.config.getint("fine_map", "x", fallback=0),
                self.config.getint("fine_map", "y", fallback=0),
                self.config.getint("fine_map", "x_off", fallback=0),
                self.config.getint("fine_map", "y_off", fallback=0),
            )
            self.coarse_map.set_dimensions(
                self.config.get("coarse_map", "path"),
                self.config.getint("coarse_map", "x", fallback=0),
                self.config.getint("coarse_map", "y", fallback=0),
                self.config.getint("coarse_map", "x_off", fallback=0),
                self.config.getint("coarse_map", "y_off", fallback=0),
            )
            self.coarse_scale = self.config.getfloat("coarse_map", "scale", fallback=1.0)
            i = 0
            while True:
                tmp_fine = "historical_fine{}".format(i)
                tmp_coarse = "historical_coarse{}".format(i)
                if not self.config.has_section(tmp_fine):
                    break
                self.historical_fine_list.append(self.config.get(tmp_fine, "path"))
                self.historical_coarse_list.append(self.config.get(tmp_coarse, "path"))
                self.rates_list.append(self.config.getfloat(tmp_fine, "rate"))
                self.times_list.append(self.config.getfloat(tmp_fine, "time"))
                i += 1
            if self.config.has_section("grid_map"):
                self.grid.file_name = "set"
                self.grid.x_size = self.config.getint("grid_map", "x")
                self.grid.y_size = self.config.getint("grid_map", "y")
                self.sample_map.x_offset = self.config.getint("sample_grid", "x_off")
                self.sample_map.y_offset = self.config.getint("sample_grid", "y_off")
            if self.config.has_section("death"):
                self.death_map.file_name = self.config.get("death", "map")
            if self.config.has_section("reproduction"):
                self.reproduction_map.file_name = self.config.get("reproduction", "map")
            if self.config.has_section("dispersal"):
                self.dispersal_method = self.config.get("dispersal", "method")
                self.m_prob = self.config.getfloat("dispersal", "m_probability", fallback=0.0)
                self.cutoff = self.config.getfloat("dispersal", "cutoff", fallback=0.0)
                self.restrict_self = self.config.getboolean("dispersal", "restrict_self", fallback=False)
                self.landscape_type = self.config.get("dispersal", "landscape_type", fallback="closed")
                self.dispersal_map = self.config.get("dispersal", "dispersal_file", fallback=None)

    def write_config(self, config_file):
        """
        Writes the config to the config file provided, overwriting any existing config files.

        :param config_file: the config file to write out to

        :rtype: None
        """
        if not self.config.has_section("main"):
            self.create_config()
            self.create_map_config()
        if config_file is not None:
            check_parent(config_file)
            if os.path.exists(config_file) and not self.config_open:
                os.remove(config_file)
        with open(config_file, "w") as f:
            self.config.write(f)
        self.config_open = True
        self.full_config_file = config_file

    def create_temporal_sampling_config(self):
        """
        Creates the time-sampling config file.

        Function is called automatically when creating a config file, and should not be manually called.
        """
        if self.times and len(self.times) != 0:
            self.config.add_section("times")
            for i, j in enumerate(sorted(set(self.times))):
                self.config.set("times", "time{}".format(i), str(j))
            if self.full_config_file is not None:
                self.write_config(self.full_config_file)

    def create_map_config(self, output_file=None):
        """
        Generates the map config file from reading the spatial structure of each of the provided files.

        :param str output_file: the file to output configuration data to (the map config file)
        """
        self.set_config_file(output_file)
        if self.is_setup_map:
            if self.is_spatial:
                self.config.add_section("sample_grid")
                if self.sample_map.file_name is None:
                    self.config.set("sample_grid", "path", "null")
                else:
                    self.config.set("sample_grid", "path", self.sample_map.file_name)
                x, y = map(str, self.sample_map.get_x_y())
                self.config.set("sample_grid", "x", str(x))
                self.config.set("sample_grid", "y", str(y))
                self.config.set(
                    "sample_grid",
                    "uses_spatial_sampling",
                    str(int(self.uses_spatial_sampling)),
                )
                self.config.add_section("fine_map")
                self.config.set("fine_map", "path", self.fine_map.file_name)
                x, y = map(str, self.fine_map.get_x_y())
                self.config.set("fine_map", "x", str(x))
                self.config.set("fine_map", "y", str(y))
                self.config.set("fine_map", "x_off", str(self.fine_map.x_offset))
                self.config.set("fine_map", "y_off", str(self.fine_map.y_offset))
                self.config.add_section("coarse_map")
                if self.coarse_map.file_name is not None:
                    self.config.set("coarse_map", "path", self.coarse_map.file_name)
                    x, y = map(str, self.coarse_map.get_x_y())
                    self.config.set("coarse_map", "x", str(x))
                    self.config.set("coarse_map", "y", str(y))
                    self.config.set("coarse_map", "x_off", str(self.coarse_map.x_offset))
                    self.config.set("coarse_map", "y_off", str(self.coarse_map.y_offset))
                    self.config.set("coarse_map", "scale", str(self.coarse_scale))
                else:
                    self.config.set("coarse_map", "path", "none")
                    self.config.set("coarse_map", "x", "0")
                    self.config.set("coarse_map", "y", "0")
                    self.config.set("coarse_map", "x_off", "0")
                    self.config.set("coarse_map", "y_off", "0")
                    self.config.set("coarse_map", "scale", "1")
                self.sort_historical_maps()
                if len(self.historical_coarse_list) != 0 and len(self.historical_fine_list) != 0:
                    for i, t in enumerate(self.times_list):
                        tmp_fine = "historical_fine{}".format(i)
                        self.config.add_section(tmp_fine)
                        self.config.set(tmp_fine, "path", self.historical_fine_list[i])
                        self.config.set(tmp_fine, "number", str(i))
                        self.config.set(tmp_fine, "time", str(t))
                        self.config.set(tmp_fine, "rate", str(self.rates_list[i]))
                        tmp_coarse = "historical_coarse{}".format(i)
                        self.config.add_section(tmp_coarse)
                        self.config.set(tmp_coarse, "path", self.historical_coarse_list[i])
                        self.config.set(tmp_coarse, "number", str(i))
                        self.config.set(tmp_coarse, "time", str(t))
                        self.config.set(tmp_coarse, "rate", str(self.rates_list[i]))
                if self.grid.file_name == "set":
                    self.config.add_section("grid_map")
                    self.config.set("grid_map", "x", str(self.grid.x_size))
                    self.config.set("grid_map", "y", str(self.grid.y_size))
                    self.config.set("sample_grid", "x_off", str(self.sample_map.x_offset))
                    self.config.set("sample_grid", "y_off", str(self.sample_map.y_offset))
                if self.death_map.file_name not in [None, "none", "null"]:
                    self.config.add_section("death")
                    self.config.set("death", "map", self.death_map.file_name)
                if self.reproduction_map.file_name not in [None, "none", "null"]:
                    self.config.add_section("reproduction")
                    self.config.set("reproduction", "map", self.reproduction_map.file_name)
                if self.dispersal_method is not None:
                    self.config.add_section("dispersal")
                    self.config.set("dispersal", "method", self.dispersal_method)
                    self.config.set("dispersal", "m_probability", str(self.m_prob))
                    self.config.set("dispersal", "cutoff", str(self.cutoff))
                    if self.restrict_self:
                        self.config.set("dispersal", "restrict_self", str(int(self.restrict_self)))
                if self.landscape_type != "closed":
                    if not self.config.has_section("dispersal"):
                        self.config.add_section("dispersal")
                    self.config.set("dispersal", "landscape_type", self.landscape_type)
                if self.dispersal_map.file_name not in ["none", None]:
                    if not self.config.has_section("dispersal"):
                        self.config.add_section("dispersal")
                    self.config.set("dispersal", "dispersal_file", self.dispersal_map.file_name)
                if self.full_config_file is not None:
                    self.write_config(self.full_config_file)
        else:
            raise RuntimeError("Cannot generate map config file without setting up map variables")

    def create_config(self, output_file=None):
        """
        Generates the configuration. This will be written out either by providing an output file here, or by calling
        write_config();

        :param str output_file: the file to generate the config option. Must be a path to a .txt file.
        """
        self.set_config_file(output_file)
        if self.is_setup_map and self.is_setup_param:
            # New method using ConfigParser
            self.config.add_section("main")
            self.config.set("main", "seed", str(self.seed))
            self.config.set("main", "task", str(self.task))
            self.config.set("main", "output_directory", self.output_directory)
            self.config.set("main", "min_spec_rate", str(self.min_speciation_rate))
            if self.is_spatial:
                self.config.set("main", "sigma", str(self.sigma))
                self.config.set("main", "tau", str(self.tau))
                self.config.set("main", "dispersal_relative_cost", str(self.dispersal_relative_cost))
            self.config.set("main", "deme", str(self.deme))
            self.config.set("main", "sample_size", str(self.sample_size))
            self.config.set("main", "max_time", str(self.max_time))
            self.config.set("main", "min_species", str(self.min_num_species))
            if self.protracted:
                self.config.add_section("protracted")
                self.config.set("protracted", "has_protracted", str(int(self.protracted)))
                self.config.set(
                    "protracted",
                    "min_speciation_gen",
                    str(float(self.min_speciation_gen)),
                )
                self.config.set(
                    "protracted",
                    "max_speciation_gen",
                    str(float(self.max_speciation_gen)),
                )
            if len(self.speciation_rates) != 0:
                self.config.add_section("spec_rates")
                for i, j in enumerate(set([x for x in self.speciation_rates])):
                    spec_rate = "spec_{}".format(i)
                    self.config.set("spec_rates", spec_rate, str(j))
            if self.full_config_file is not None:
                self.write_config(self.full_config_file)
        else:
            raise RuntimeError("Setup has not been completed, cannot create config file")
        self.create_temporal_sampling_config()

    def run_simple(self, seed, task, output, speciation_rate, sigma, size):
        """
        Runs a simple coalescence simulation on a square infinite landscape with the provided parameters.
        This requires a separate compilation of the inf_land version of the coalescence simulator.

        Note that this function returns richness=0 for failure to read from the file. It is assumed that there will
        be at least one species in the simulation.

        Note that the maximum time for this function is set as 10 hours (36000 seconds) and will raise an exception if
        the simulation does not complete in this time).

        :raises RuntimeError: if the simulation didn't complete in time.

        :param seed: the simulation seed
        :param task: the task (for file naming)
        :param output: the output directory
        :param speciation_rate: the probability of speciation
        :param sigma: the normal distribution sigma value for dispersal
        :param size: the size of the world (so there will be size^2 individuals simulated)

        :return: the species richness in the simulation
        """
        self.set_simulation_parameters(
            seed=seed,
            task=task,
            output_directory=output,
            min_speciation_rate=speciation_rate,
            sigma=sigma,
            tau=1,
            deme=1.0,
            sample_size=1.0,
            max_time=36000,
            dispersal_relative_cost=1,
            min_num_species=1,
        )
        self.set_map("null", size, size)
        self.set_speciation_rates([speciation_rate])
        self.run()
        # Now read the species richness from the database
        try:
            richness = self.get_species_richness()
            if richness == 0:
                raise AssertionError(
                    "Richness is 0, error in program. This should never be the case for a complete " "sim."
                )
            return richness
        except IOError:
            raise RuntimeError(
                "Simulation didn't complete in 10 hours (maximum time for run_simple). Try running a "
                "custom simulation instead."
            )

    def get_species_richness(self, reference=1):
        """
        Calls coal_analyse.get_species_richness() with the supplied variables.

        Requires successful import of coal_analyse and sqlite3.

        :param int reference: the community reference to obtain the metrics for.

        :return: the species richness.
        """
        self.calculate_sql_database()
        db = self.output_database
        t = CoalescenceTree()
        t.set_database(db)
        return t.get_species_richness(reference)

    def get_protracted(self):
        """
        Gets whether the simulation pointed to by this object is a protracted simulation or not.
        """
        if self.output_database is None or not os.path.exists(self.output_database):
            return self.protracted
        else:
            t = CoalescenceTree()
            try:
                t.set_database(self.output_database)
            except IOError:
                # Catches errors thrown if the simulation is paused
                pass
            return t.is_protracted()

    def set_map_files(
        self,
        sample_file,
        fine_file=None,
        coarse_file=None,
        historical_fine_file=None,
        historical_coarse_file=None,
        dispersal_map=None,
        death_map=None,
        reproduction_map=None,
    ):
        """
        Sets the map files (or to null, if none specified). It then calls detect_map_dimensions() to correctly read in
        the specified dimensions.

        If sample_file is "null", dimension values will remain at 0.
        If coarse_file is "null", it will default to the size of fine_file with zero offset.
        If the coarse file is "none", it will not be used.
        If the historical fine or coarse files are "none", they will not be used.

        .. note:: the dispersal map should be of dimensions xy by xy where x, y are the fine map dimensions. Dispersal
                  rates from each row/column index represents dispersal from the
                  row index to the column index according to index = x+(y*xdim), where x,y are the coordinates of the
                  cell and xdim is the x dimension of the fine map. See the
                  :class:`PatchedLandscape class <pycoalescence.patched_landscape.PatchedLandscape>` for routines for
                  generating these landscapes.

        :param str sample_file: the sample map file. Provide "null" if on samplemask is required
        :param str fine_file: the fine map file. Defaults to "null" if none provided
        :param str coarse_file: the coarse map file. Defaults to "none" if none provided
        :param str historical_fine_file: the historical fine map file. Defaults to "none" if none provided
        :param str historical_coarse_file: the historical coarse map file. Defaults to "none" if none provided
        :param str dispersal_map: the dispersal map for reading dispersal values. Default to "none" if none provided
        :param str death_map: a map of relative death probabilities, at the scale of the fine map
        :param str reproduction_map: a map of relative reproduction probabilities, at the scale of the fine map

        :rtype: None

        :return: None
        """
        if dispersal_map is None:
            self.dispersal_map.file_name = "none"
        else:
            self.dispersal_map.file_name = dispersal_map
        if death_map is None:
            self.death_map.file_name = "none"
        else:
            self.death_map.file_name = death_map
        if reproduction_map is None:
            self.reproduction_map.file_name = "none"
        else:
            self.reproduction_map.file_name = reproduction_map
        Landscape.set_map_files(
            self,
            sample_file,
            fine_file,
            coarse_file,
            historical_fine_file,
            historical_coarse_file,
        )

    def detect_map_dimensions(self):
        """
        Detects all the map dimensions for the provided files (where possible) and sets the respective values.
        This is intended to be run after set_map_files()

        :return: None
        """
        Landscape.detect_map_dimensions(self)
        # Now do detection for dispersal map
        self.check_dispersal_map()
        self.check_reproduction_map()
        self.check_death_map()
        self.check_maps()

    def check_dispersal_map(self):
        """Checks that the dispersal map dimensions match the fine map dimensions."""
        if self.dispersal_map.file_name not in {"none", "null", None}:
            if self.coarse_map.file_name not in {None, "none"}:
                raise ValueError(
                    "Cannot use a coarse map if using a dispersal map. " "This feature is currently unsupported."
                )
            self.dispersal_map.set_dimensions()
            if (
                self.dispersal_map.x_size != self.fine_map.x_size * self.fine_map.y_size
                or self.dispersal_map.y_size != self.fine_map.x_size * self.fine_map.y_size
            ):
                raise ValueError(
                    "Dimensions of dispersal map do not match dimensions of fine map. This is currently" " unsupported."
                )

    def check_reproduction_map(self):
        """Checks that the reproduction map dimensions match the fine map dimensions."""
        if self.reproduction_map.file_name not in {None, "none", "null"}:
            if self.coarse_map.file_name not in {None, "none", "null"}:
                raise ValueError(
                    "Cannot use a coarse map if using a reproduction map. " "This feature is currently unsupported."
                )
        self.check_dimensions_match_fine(self.reproduction_map, "reproduction")

    def check_death_map(self):
        """Checks that the death map dimensions match the fine map dimensions."""
        if self.death_map.file_name not in {None, "none", "null"}:
            if self.coarse_map.file_name not in {None, "none", "null"}:
                raise ValueError(
                    "Cannot use a coarse map if using a death map. " "This feature is currently unsupported."
                )
        self.check_dimensions_match_fine(self.death_map, "death")

    def add_dispersal_map(self, dispersal_map):
        """
        Adds a dispersal map to the simulation.

        :param str/pycoalescence.Map dispersal_map: the dispersal map to import

        :rtype: None
        """
        if isinstance(dispersal_map, Map):
            self.dispersal_map = dispersal_map
        else:
            self.dispersal_map.file_name = dispersal_map
        self.check_dispersal_map()

    def add_death_map(self, death_map):
        """
        Adds a death map to the simulation.

        :param str/pycoalescence.Map death_map: the death map to import

        :rtype: None
        """
        if isinstance(death_map, Map):
            self.death_map = death_map
        else:
            self.death_map.file_name = death_map
        self.check_death_map()

    def add_reproduction_map(self, reproduction_map):
        """
        Adds a death map to the simulation.

        :param str/pycoalescence.Map reproduction_map: the death map to import

        :rtype: None
        """
        if isinstance(reproduction_map, Map):
            self.reproduction_map = reproduction_map
        else:
            self.reproduction_map.file_name = reproduction_map
        self.check_reproduction_map()

    def add_gillespie(self, generations=0.0):
        """
        Uses the Gillespie algorithm from the given number of generations into the simulation.
        :param generations: the number of generations at which to use gillespie.

        :rtype: None
        :return: None
        """
        FutureWarning(
            "The gillespie algorithm is not implemented beyond the base functionality and is awaiting a future update."
        )
        if generations < 0:
            raise ValueError("Number of generations cannot be < 0.")
        if not self.check_can_use_gillespie():
            self.uses_gillespie = False
            raise ValueError(
                "Cannot use Gillespie with this simulation. Gillespie algorithm can only be used for spatial "
                "simulations involving a dispersal map and point-speciation. Ensure your simulation is correctly set up."
            )
        self.uses_gillespie = True
        self.gillespie_threshold = float(generations)

    def check_can_use_gillespie(self):
        """
        Checks if the simulation can use the Gillespie algorithm.

        :return: true if the simulation can use Gillespie
        :rtype: bool
        """
        if not (self.is_setup_map and self.is_setup_param):
            raise RuntimeError("Cannot check if Gillespie is possible before simulation is setup.")
        if (
            self.dispersal_map.file_name in ["none", None]
            or not self.is_spatial
            or self.coarse_map.file_name not in ["none", None]
            or self.protracted
            or self.deme < 1
        ):
            return False
        return True

    def check_dimensions_match_fine(self, map_to_check, name=""):
        """
        Checks that the dimensions of the provided map matches the fine map.

        :param Map map_to_check: map to check the dimensions of against the fine map
        :param str name: name to write out in error message

        :return: true if the dimensions match
        """
        if map_to_check.file_name not in {"none", "null", None}:
            map_to_check.set_dimensions()
            if not map_to_check.has_equal_dimensions(self.fine_map):
                # if the sizes match, then proceed with a warning
                if map_to_check.x_size == self.fine_map.x_size and map_to_check.y_size == self.fine_map.y_size:
                    self.logger.warning("Coordinates of {} map did not match fine map.".format(name))
                    map_to_check.x_offset = self.fine_map.x_offset
                    map_to_check.y_offset = self.fine_map.y_offset
                    map_to_check.x_res = self.fine_map.x_res
                    map_to_check.y_res = self.fine_map.y_res
                else:
                    raise ValueError(
                        "Dimensions of the {} map do not match the fine map. This is currently "
                        "unsupported.".format(name)
                    )

    def set_simulation_parameters(
        self,
        seed,
        task,
        output_directory,
        min_speciation_rate,
        sigma=1.0,
        tau=1.0,
        deme=1.0,
        sample_size=1.0,
        max_time=3600,
        dispersal_method=None,
        m_prob=0.0,
        cutoff=0,
        dispersal_relative_cost=1,
        min_num_species=1,
        restrict_self=False,
        landscape_type=False,
        protracted=False,
        min_speciation_gen=None,
        max_speciation_gen=None,
        spatial=True,
        uses_spatial_sampling=False,
        times=None,
    ):
        """
        Set all the simulation parameters apart from the map objects.

        :param int seed: the unique job number for this simulation set
        :param int task: the task reference number (used for easy file identification after simulations are complete)
        :param str output_directory: the output directory to store the SQL database
        :param float min_speciation_rate: the minimum speciation rate to simulate
        :param float sigma: the dispersal sigma value
        :param float tau: the fat-tailed dispersal tau value
        :param float deme: the deme size (in individuals per cell)
        :param float sample_size: the sample size of the deme (decimal 0-1)
        :param float max_time: the maximum allowed simulation time (in seconds)
        :param str dispersal_method: the dispersal kernel method. Should be one of [normal, fat-tail, norm-uniform]
        :param float m_prob: the probability of drawing from the uniform dispersal. Only relevant for uniform dispersals
        :param float cutoff: the maximum value for the uniform dispersal. Only relevant for uniform dispersals.
        :param float dispersal_relative_cost: the relative cost of travelling through non-habitat (defaults to 1)
        :param int min_num_species: the minimum number of species known to exist (defaults to 1
        :param bool restrict_self: if true, restricts dispersal from own cell
        :param bool/str landscape_type: if false or "closed", restricts dispersal to the provided maps, otherwise
                                        can be "infinite", or a tiled landscape using "tiled_coarse" or "tiled_fine",
                                        or a clamped landscape using "clamped_coarse" or "clamped_fine".
        :param bool protracted: if true, uses protracted speciation application
        :param float min_speciation_gen: the minimum amount of time a lineage must exist before speciation occurs.
        :param float max_speciation_gen: the maximum amount of time a lineage can exist before speciating.
        :param bool spatial: if true, means that the simulation is spatial
        :param bool uses_spatial_sampling: if true, the sample mask is interpreted as a proportional sampling mask,
                                           where the number of individuals sampled in the cell is equal to the
                                           density * deme_sample * cell sampling proportion
        :param list times: list of temporal sampling points to apply (in generations)
        """
        if not self.is_setup_param:
            self.set_seed(seed)
            self.output_directory = output_directory
            self.task = task
            self.min_speciation_rate = min_speciation_rate
            self.sigma = sigma
            self.tau = tau
            self.deme = deme
            self.sample_size = sample_size
            self.max_time = max_time
            self.dispersal_relative_cost = dispersal_relative_cost
            self.min_num_species = min_num_species
            self.dispersal_method = dispersal_method
            self.m_prob = m_prob
            self.cutoff = cutoff
            self.calculate_sql_database()
            self.is_setup_param = True
            self.restrict_self = restrict_self
            self.is_spatial = spatial
            if uses_spatial_sampling and not spatial:
                raise ValueError("Must use a spatial simulation to define spatial sampling regime.")
            self.uses_spatial_sampling = uses_spatial_sampling
            if not spatial:
                self.is_setup_map = True
            if landscape_type in {False, "closed"}:
                self.landscape_type = "closed"
            elif landscape_type in {True, "infinite"}:
                self.landscape_type = "infinite"
            elif landscape_type in {
                "tiled_coarse",
                "tiled_fine",
                "clamped_coarse",
                "clamped_fine",
            }:
                self.landscape_type = landscape_type
            else:
                raise ValueError("Supplied landscape type is not recognised: {}".format(landscape_type))
            if protracted:
                self.protracted = True
                if max_speciation_gen is None:
                    self.logger.warning(
                        "Using protracted speciation, but no maximum speciation generation supplied."
                        " Default to 10^100"
                    )
                    self.max_speciation_gen = 10 ** 100
                else:
                    self.max_speciation_gen = max_speciation_gen
                if min_speciation_gen is None:
                    self.logger.warning(
                        "Using protracted speciation, but no minimum speciation generation supplied." " Default to 0.0"
                    )
                    self.min_speciation_gen = 0.0
                else:
                    self.min_speciation_gen = min_speciation_gen
            if times:
                self.add_sample_time(times)
        else:
            self.logger.warning("Parameters already set up.")

    def set_seed(self, seed):
        """
        Sets the seed for the simulation.

        A seed < 1 should not be set for the necsim, as equivalent behaviour is produce for seed and abs(seed), plus for
        seed = 1 and seed = 0. Consequently, for any values less than 1, we take a very large number plus the seed,
        instead. Therefore a error is raised if the seed exceeds this very large number (this is an acceptable
        decrease in userability as a seed that large is unlikely to ever be used).

        :param int seed: the random number seed
        """
        limit_val = int(2147483647 / 2)
        if seed > limit_val:
            raise ValueError("Seed cannot be larger than 2,147,483,647")
        if seed < 1:
            new_seed = abs(seed) + limit_val
            self.logger.critical("Seed of {} is < 1, so will be changed to {}\n".format(seed, new_seed))
            self.seed = new_seed
        else:
            self.seed = seed

    def check_simulation_parameters(self, expected=False, ignore_errors=False):
        """
        Checks that simulation parameters have been correctly set and the program is ready for running.

        :param ignore_errors: if true, any FileNotFoundError and FileExistsError raised by checking the output database
                              are ignored
        :param expected: set to true if we expect the output file to exist
        """
        if not self.is_setup_param or (self.seed == 0 and self.task == 0):
            raise RuntimeError("Simulation parameters have not been set.")
        if self.output_directory in {"", None, "null"}:
            raise RuntimeError("Output directory not set.")
        if self.full_config_file is not None:
            check_parent(self.full_config_file)
        if self.uses_gillespie and not self.check_can_use_gillespie():
            raise RuntimeError("Cannot use Gillespie algorithm with current map setup.")
        try:
            if self.output_database is None:
                self.calculate_sql_database()
            self.check_file_parameters(expected=expected)
        except (FileExistsError, FileNotFoundError) as err:  # pragma: no cover
            if not ignore_errors:
                raise err
            else:
                self.logger.info(str(err))

    def resume_coalescence(
        self,
        pause_directory,
        seed,
        task,
        max_time,
        out_directory=None,
        protracted=None,
        spatial=None,
    ):
        """
        Resumes the simulation from the specified directory, looking for the simulation with the specified seed and task
        referencing.

        :param pause_directory: the directory to search for the paused simulation
        :param seed: the seed of the paused simulation
        :param task: the task of the paused simulation
        :param max_time: the maximum time to run simulations for
        :param out_directory: optionally provide an alternative output location. Defaults to same location as
        pause_directory
        :param bool protracted: protractedness of the simulation
        :param bool spatial: if the simulation is to be run with spatial complexity

        .. note: A max time of 0 uses the previous simulation maximum time.

        :return: None
        """
        if protracted is not None:
            self.protracted = protracted
        if spatial is not None:
            self.is_spatial = spatial
        if out_directory is None:
            out_directory = pause_directory
        self.output_directory = out_directory
        self.pause_directory = pause_directory
        self.seed = seed
        self.task = task
        self.max_time = max_time
        self.calculate_sql_database()
        file_path = os.path.join(pause_directory, "Pause", "Dump_main_{}_{}.csv".format(task, seed))
        if not os.path.exists(file_path):
            raise IOError(
                "Paused file " + file_path + " not found. Ensure pause directory is correct and is accessible."
            )
        self.setup_necsim()
        self.c_simulation.setup_resume(pause_directory, out_directory, int(seed), int(task), int(max_time))
        self.is_setup_complete = True
        self._run_and_output()

    def persistent_ram_usage(self):
        """
        This is the persistent RAM usage which cannot be optimised by the program for a particular set of maps

        :return: the total persistent RAM usage in bytes
        """
        total_ram = 0
        total_ram += self.fine_map.x_size * self.fine_map.y_size * 4
        # First add the maps
        if self.coarse_map.file_name not in [None, "none"]:
            total_ram += self.coarse_map.x_size * self.coarse_map.y_size * 4
        if len(self.historical_fine_list) != 0:
            total_ram += self.fine_map.x_size * self.fine_map.y_size * 4
        if len(self.historical_coarse_list) != 0:
            total_ram += self.coarse_map.x_size * self.coarse_map.y_size * 4
        if self.death_map.file_name not in [None, "none"]:
            total_ram += self.death_map.x_size * self.death_map.y_size * 8
        if self.sample_map.file_name not in [None, "none", "null"]:
            total_ram += self.sample_map.x_size * self.sample_map.y_size * 1
        if self.dispersal_map.file_name not in [None, "none"]:
            total_ram += self.fine_map.x_size * self.fine_map.y_size * 8
        # Now add the other large memory objects
        # The data object
        total_ram += 2 * self.count_individuals() * 91 * min(len(self.times_list), 1)
        # The active object
        total_ram += 72 * self.count_individuals() * min(len(self.times_list), 1)
        return total_ram

    def count_individuals(self):
        """
        Estimates the number of individuals to be simulated. This may be inaccurate if using multiple time points and
        historical maps.

        :return: a count of the number of individuals to be simulated

        :rtype: float
        """
        if self.count_total is None:
            if self.sample_map.file_name in [None, "none", "null"]:
                ds_array = np.ones(shape=(self.fine_map.x_size, self.fine_map.y_size))
            else:
                ds = gdal.Open(self.sample_map.file_name)
                ds_array = ds.GetRasterBand(1).ReadAsArray()
            self.count_total = np.sum(ds_array) * self.sample_size * self.get_average_density()
            if len(self.times_list) > 0:
                self.count_total *= len(self.times_list)
            return self.count_total
        else:
            return self.count_total

    def import_fine_map_array(self):
        """
        Imports the fine map array to the in-memory object, subsetted to the same size as the sample grid.

        :rtype: None
        """
        if self.fine_map_array is None:
            if self.fine_map.file_name == "null":
                self.fine_map_array = np.ones((self.fine_map.x_size, self.fine_map.y_size))
            else:
                self.fine_map_array = self.fine_map.get_subset(
                    self.fine_map.x_offset,
                    self.fine_map.y_offset,
                    self.sample_map.x_size,
                    self.sample_map.y_size,
                )

    def import_sample_map_array(self):
        """
        Imports the sample map array to the in-memory object.

        :rtype: None
        """
        if self.sample_map_array is None:
            if self.sample_map.file_name == "null":
                self.sample_map_array = np.ones((self.sample_map.x_size, self.sample_map.y_size))
            else:
                self.sample_map.open()
                if self.uses_spatial_sampling:
                    self.sample_map_array = self.sample_map.data
                else:
                    self.sample_map_array = np.ma.masked_where(self.sample_map.data >= 0.5, self.sample_map.data).mask
                self.sample_map.data = None

    def grid_density_estimate(self, x_off, y_off, x_dim, y_dim):
        """
        Counts the density total for a subset of the grid by sampling from the fine map

        :note: This is an approximation (based on the average density of the fine map) and does not produce
               a perfect value. This is done for performance reasons. The actual value can be obtained with
               grid_density_actual().

        :param x_off: the x offset of the grid map subset
        :param y_off: the y offset of the grid map subset
        :param x_dim: the x dimension of the grid map subset
        :param y_dim: the y dimension of the grid map subset

        :return: an estimate of the total individuals that exist in the subset.

        :rtype: int
        """
        self.import_fine_map_array()
        if (
            self.sample_map.file_name not in ["none", "null", None]
            and self.sample_map.file_name != self.fine_map.file_name
        ):
            self.import_sample_map_array()
            # Less accurate, but faster way.
            arr_subset = self.sample_map_array[y_off : y_off + y_dim, x_off : x_off + x_dim]
            return int(np.sum(np.floor(arr_subset * self.get_average_density() * self.sample_size)))
        return int(np.sum(np.floor(self.fine_map_array * self.deme * self.sample_size)))

    def grid_density_actual(self, x_off, y_off, x_dim, y_dim):
        """
        Counts the density total for a subset of the grid by sampling from the fine map.

        Note that for large maps this can take a very long time.

        :param x_off: the x offset of the grid map subset
        :param y_off: the y offset of the grid map subset
        :param x_dim: the x dimension of the grid map subset
        :param y_dim: the y dimension of the grid map subset

        :return: the total individuals that exist in the subset.

        :rtype: int
        """
        self.import_fine_map_array()
        if (
            self.sample_map.file_name not in ["none", "null", None]
            and self.sample_map.file_name != self.fine_map.file_name
        ):
            self.import_sample_map_array()
            # Less accurate, but faster way.
            return int(
                np.sum(
                    np.floor(
                        np.multiply(
                            self.fine_map_array[y_off : y_off + y_dim, x_off : x_off + x_dim],
                            self.sample_map_array[y_off : y_off + y_dim, x_off : x_off + x_dim],
                        )
                        * self.deme
                        * self.sample_size
                    )
                )
            )
        else:
            return int(
                np.sum(
                    np.floor(
                        self.fine_map_array[y_off : y_off + y_dim, x_off : x_off + x_dim] * self.deme * self.sample_size
                    )
                )
            )

    def get_average_density(self):
        """
        Gets the average density across the fine map, subsetted for the sample grid.
        """
        if self.fine_map_average is None:
            self.import_fine_map_array()
            self.fine_map_average = self.deme * np.mean(self.fine_map_array)
        return self.fine_map_average

    def check_sample_map_equals_sample_grid(self):
        """
        Checks if the grid and sample map are the same size and offset (in which case, future operations can be
        simplified).

        :return: true if the grid and sample map dimensions and offsets are equal
        :rtype: bool
        """
        return (
            self.grid.x_size == self.sample_map.x_size
            and self.grid.y_size == self.sample_map.x_size
            and self.grid.x_offset == 0
            and self.grid.y_offset == 0
        )

    def optimise_ram(self, ram_limit):
        """
        Optimises the maps for a specific RAM usage.

        If ram_limit is None, this function does nothing.

        :note: Assumes that the C++ compiler has sizeof(long) = 8 bytes for calculating space usage.

        :note: Only optimises RAM for a square area of the map. For rectangular shapes, will use the shortest length as
               a maximum size.

        :param ram_limit: the desired amount of RAM to limit to, in GB

        :raises MemoryError: if the desired simulation cannot be compressed into available RAM
        """
        if self.sample_size <= 0 or self.deme <= 0:
            raise ValueError("Sample size is 0, or deme is 0. set_simulation_parameters() before optimising RAM.")
        self.grid = copy.deepcopy(self.sample_map)
        # Over-estimate static usage slightly
        static_usage = 1.1 * self.persistent_ram_usage() / 1024 ** 3
        if static_usage > ram_limit:
            raise MemoryError("Cannot achieve RAM limit: minimum requirements are {}GB.".format(round(static_usage, 2)))
        remaining_space = (ram_limit - static_usage) * 1024 ** 3
        self.logger.info("Remaining space is {}GB.\n".format(round(remaining_space / 1024 ** 3, 2)))
        # Rough calculation of number of cells we have remaining (aim for 75% to be sure)
        average_density = self.get_average_density()
        # This is the size of the grid object in memory
        tmp_space = 0.95 * (remaining_space / ((16 * average_density) + 32))
        size = int(np.rint(tmp_space ** 0.5))
        if size < self.sample_map.x_size or size < self.sample_map.y_size:
            if size > self.sample_map.x_size or size > self.sample_map.y_size:  # pragma: no cover
                size = min(self.sample_map.x_size, self.sample_map.y_size)
            if size < 2:
                raise MemoryError("Could not find square grid to achieve RAM limit. Please set this manually.")
            max_attained = 0
            max_x = 0
            max_y = 0
            # Now loop over every square of the required size on the grid, until we find one with the maximum density
            if (self.sample_map.x_size * self.sample_map.y_size) > 1000000:  # pragma: no cover
                printing = True
            else:
                printing = False
            end_x = self.sample_map.x_size - size - 1
            if size < 10:  # pragma: no cover
                self.logger.warning("Extremely small grid size: {}. Disabling maximum density checking.".format(size))
                max_x = int(self.sample_map.x_size / 2)
                max_y = int(self.sample_map.y_size / 2)
                max_attained = 1
            else:
                for x in range(
                    0,
                    max(self.sample_map.x_size - size - 1, 1),
                    int(round(max(10.0, size / 10))),
                ):
                    if printing:  # pragma: no cover
                        self.logger.info("\rChecking {} / {} : {}%".format(x, end_x, round(100 * x / end_x)))
                    for y in range(
                        0,
                        max(self.sample_map.y_size - size - 1, 1),
                        int(round(max(10.0, size / 10))),
                    ):
                        # print("x, y: {}, {}".format(x, y))
                        tmp_total = self.grid_density_actual(x, y, size, size)
                        if tmp_total > max_attained:
                            max_x = x
                            max_y = y
                            max_attained = tmp_total
                if printing:  # pragma: no cover
                    self.logger.info("\rChecking complete               \n")
            if max_attained > 0:
                # Now do one last check for consistency
                while self._fine_map_check(max_x, max_y, size):
                    if printing:  # pragma: no cover
                        printing = False
                        self.logger.info("Shrinking due to high densities in sample zone.\n")
                    # Brute-force dividing by 2 each time
                    size = max(1, int(size / 2))
                    if size == 1:
                        break
                if max_x + size > self.sample_map.x_size or max_y + size > self.sample_map.y_size:  # pragma: no cover
                    self.logger.debug("Max x,y: {}, {}".format(max_x, max_y))
                    self.logger.debug("Size: {}".format(size))
                    self.logger.debug("Fine map size: {}, {}".format(self.fine_map.x_size, self.fine_map.y_size))
                    self.logger.debug("Sample map size: {}, {}".format(self.sample_map.x_size, self.sample_map.y_size))
                    raise RuntimeError("Incorrect dimension setting - please report this bug")
                self.grid.x_size = size
                self.grid.y_size = size
                self.sample_map.x_offset = max_x
                self.sample_map.y_offset = max_y
                self.grid.file_name = "set"
                if self.check_sample_map_equals_sample_grid():  # pragma: no cover
                    self.grid.file_name = "none"
            else:  # pragma: no cover
                raise MemoryError(
                    "Could not optimise maps for desired memory usage. "
                    "Attempted grid size of {} and achived max of {}".format(size, max_attained)
                )
        self._wipe_objects()

    def get_optimised_solution(self):
        """
        Gets the optimised solution as a dictionary containing the important optimised variables.
        This can be read back in with set_optimised_solution

        :return: dict containing the important optimised variables
        :rtype: dict
        """
        return {
            "grid_x_size": self.grid.x_size,
            "grid_y_size": self.grid.y_size,
            "sample_x_offset": self.sample_map.x_offset,
            "sample_y_offset": self.sample_map.y_offset,
            "grid_file_name": self.grid.file_name,
        }.copy()

    def set_optimised_solution(self, dict_in):
        """
        Sets the optimised RAM solution from the variables in the provided dictionary.
        This should contain the grid_x_size, grid_y_size, grid_file_name, sample_x_offset and sample_y_offset.

        :param dict dict_in: the dictionary containing the optimised RAM solution variables
        :rtype: None
        """
        self.grid.x_size = dict_in["grid_x_size"]
        self.grid.y_size = dict_in["grid_y_size"]
        self.sample_map.x_offset = dict_in["sample_x_offset"]
        self.sample_map.y_offset = dict_in["sample_y_offset"]
        self.grid.file_name = dict_in["grid_file_name"]

    def finalise_setup(self, expected=False, ignore_errors=False):
        """
        Runs all setup routines to provide a complete simulation. Should be called immediately before run_coalescence()
        to ensure the simulation setup is complete.

        :param ignore_errors: if true, any FileNotFoundError and FileExistsError raised by checking the output database
                              are ignored
        :param expected: set to true if we expect the output file to exist
        """
        self.check_simulation_parameters(expected=expected, ignore_errors=ignore_errors)
        self.create_config()
        self.create_map_config()
        self.calculate_sql_database()
        try:
            from pycoalescence.__init__ import __version__

            self.logger.info("Simulation set-up under pycoalescence version {}\n".format(__version__))
        except ImportError:  # pragma: no cover
            self.logger.info("pycoalescence version not found.")
        self.setup_necsim()
        if self.full_config_file is not None:  # pragma: no cover
            self.c_simulation.import_from_config(self.full_config_file)

        else:
            self.config.write(self.config_string)
            self.c_simulation.import_from_config_string(self.config_string.getvalue())
        self.c_simulation.setup()
        if self.uses_gillespie and self.check_can_use_gillespie():
            self.c_simulation.add_gillespie(self.gillespie_threshold)
        self.is_setup_complete = True

    def setup_necsim(self):
        """
        Calculates the type of the simulation (spatial/non-spatial, protracted/non-protracted) and sets the c object
        appropriately.

        :rtype: None
        """
        if self.protracted:
            if self.is_spatial:
                self.c_simulation = CPSpatialSimulation(self.logger)
            else:
                self.c_simulation = CPNSESimulation(self.logger)
        elif self.is_spatial:
            self.c_simulation = CSpatialSimulation(self.logger)
        else:
            self.c_simulation = CNSESimulation(self.logger)

    def set_speciation_rates(self, speciation_rates):
        """Add speciation rates for analysis at the end of the simulation. This is optional

        :param list speciation_rates: a list of speciation rates to apply at the end of the simulation
        """
        self.speciation_rates.extend(speciation_rates)

    def calculate_sql_database(self):
        """
        Saves the output database location to self.output_database.
        """
        self.output_database = os.path.join(
            self.output_directory,
            "data_{}_{}.db".format(str(self.task), str(self.seed)),
        )

    def check_sql_database(self, expected=False):
        """
        Checks whether the output database exists. If the existance does not match the expected variable, raises an
        error.

        :raises FileExistsError: if the file already exists when it's not expected to
        :raises FileNotExistsError: if the file does not exist when we expect it to

        :param expected: boolean for expected existance of the output file

        :rtype: None
        """
        check_parent(self.output_database)
        if not expected:
            if os.path.exists(self.output_database):
                try:
                    t = CoalescenceTree(self.output_database)
                except IOError:
                    pass
                else:
                    raise FileExistsError("Output sql database already exists.")
        else:
            if not os.path.exists(self.output_database):
                raise FileNotFoundError("Output sql database {} not found.".format(self.output_database))

    def check_maps(self):
        """
        Checks that the maps all exist and that the file structure makes sense.

        :raises ValueError: if the grid is not within the sample map.

        :return: None
        """
        if self.is_spatial:
            Landscape.check_maps(self)
            if (
                self.grid.file_name == "set"
                and self.sample_map.x_offset + self.grid.x_size > self.sample_map.x_size
                or self.sample_map.y_offset + self.grid.y_size > self.sample_map.y_size
            ):
                raise ValueError("Grid is not within the sample map - please check offsets of sample map.")
            self.check_dispersal_map()
            self.check_reproduction_map()
            self.check_death_map()

    def check_file_parameters(self, expected=False):
        """
        Check that all the required files exist for the simulation and the output doesn't already exist.

        :param expected: set to true if we expect the output file to already exist

        :raises RuntimeError: if previous set-up routines are not complete
        """
        if self.is_setup_map and self.is_setup_param:
            self.check_maps()
        else:
            raise RuntimeError("Set up of maps or parameters is incomplete.")
        self.check_sql_database(expected=expected)

    def run(self):
        """
        Convenience function which completes setup, runs the simulation and calculates the coalescence tree for the set
        speciation rates in one step.

        :rtype: None
        """
        self.finalise_setup()
        self._run_and_output()

    def _run_and_output(self):
        """
        Runs the simulation and outputs to the database. Should only be called internally.

        :return:
        """
        if self.run_coalescence():
            self.apply_speciation_rates()
        else:
            self.logger.warning("Simulation did not complete in time specified.\n")
            self.logger.warning("Resume with extra time to continue.\n")

    def run_coalescence(self):
        """
        Attempt to run the simulation with the given simulation set-up.
        This is the main routine performing the actual simulation which will take a considerable amount of time.

        :return: True if the simulation completes successfully, False if the simulation pauses.

        :rtype: bool
        """
        if self.is_setup_complete:
            # Make the output directory if it doesn't yet exist
            if not os.path.exists(self.output_directory):
                os.makedirs(self.output_directory)
            # Call the C++ object and run the simulation
            return self.c_simulation.run()
        else:
            raise RuntimeError("Set up is incomplete.")

    def apply_speciation_rates(self, speciation_rates=None):
        """
        Applies the speciation rates to the coalescence tree and outputs to the database.

        :param speciation_rates: a list of speciation rates to apply

        :rtype: None
        """
        if speciation_rates:
            for each in speciation_rates:
                self.speciation_rates.append(each)
            self.speciation_rates = [x for x in set(self.speciation_rates)]
        self.c_simulation.apply_speciation_rates(self.speciation_rates)

    def _wipe_objects(self):
        """
        Clears the arrays from memory
        :return:
        """
        self.fine_map_array = None
        self.sample_map_array = None

    def _fine_map_check(self, max_x, max_y, size):
        """
        Checks the total number of individuals in a subset from the fine map is less than the average density multiplied
        by the size of the square.

        :param max_x: the x value to start subsetting from
        :param max_y: the y value to start subsetting from
        :param size: the size of the square
        :return: bool true if it is smaller
        :rtype: bool
        """
        self._fine_map_sum_res = self.grid_density_actual(max_x, max_y, size, size)
        return self._fine_map_sum_res > size * size * self.get_average_density()



"""Tests that the fragment config correctly handlers creation of a configuration file from a shapefile and a raster."""
import csv
import os
import unittest

from setup_tests import setUpAll, tearDownAll

from pycoalescence.fragment_config import FragmentConfigHandler, generate_fragment_csv


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestFragmentConfigHandler(unittest.TestCase):
    """
    Tests that the fragment config handler works correctly.
    """

    @classmethod
    def setUpClass(cls):
        """
        Generates the expected dictionary for the object
        """
        cls.fragment_csv = os.path.join("output", "fragment_config1.csv")
        cls.expected_dict = {
            "fragment1": {"min_x": 8, "max_x": 11, "min_y": 0, "max_y": 4, "area": 40},
            "fragment2": {"min_x": 1, "max_x": 8, "min_y": 1, "max_y": 8, "area": 10},
        }
        cls.expected_dict2 = {
            "fragment1": {"min_x": 1, "max_x": 4, "min_y": 1, "max_y": 3, "area": 40},
            "fragment2": {"min_x": 7, "max_x": 10, "min_y": 5, "max_y": 11, "area": 20},
        }

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.fragment_csv):
            os.remove(cls.fragment_csv)

    def testIncorrectParametersRaiseErrors(self):
        """
        Tests that supplying incorrect parameters causes the correct error to be raised.
        """
        f = FragmentConfigHandler()
        with self.assertRaises(ValueError):
            f.generate_config(
                input_shapefile=os.path.join("output", "not_a_shape.tif"),
                input_raster=os.path.join("output", "SA_samplemask.tif"),
            )
        with self.assertRaises(ValueError):
            f.generate_config(
                input_shapefile=os.path.join("output", "not_a_shape.shp"),
                input_raster=os.path.join("output", "SA_samplemask.tfi"),
            )
        with self.assertRaises(IOError):
            f.generate_config(
                input_shapefile=os.path.join("output", "not_a_shape.shp"),
                input_raster=os.path.join("sample", "SA_sample.tif"),
            )
        with self.assertRaises(IOError):
            f.generate_config(
                input_shapefile=os.path.join("sample", "shape_sample.shp"),
                input_raster=os.path.join("sample", "SA_samplenot.tif"),
            )
        with self.assertRaises(IOError):
            f.write_csv(output_csv=os.path.join("sample", "FragmentsTest.csv"))
        with self.assertRaises(ValueError):
            f.generate_config(
                input_shapefile=os.path.join("sample", "shape_sample.shp"),
                input_raster=os.path.join("sample", "SA_sample.tif"),
            )

    def testGenerateFragments(self):
        """
        Tests that the fragments are correctly generated from the extents on the map.
        """
        f = FragmentConfigHandler()
        f.generate_config(
            input_shapefile=os.path.join("sample", "shape_sample.shp"),
            input_raster=os.path.join("sample", "SA_sample_fine.tif"),
        )
        for each in self.expected_dict.keys():
            self.assertDictEqual(self.expected_dict[each], f.fragment_list[each])

    def testWritesCsv(self):
        """Test writes the csv correctly."""
        f = FragmentConfigHandler()
        f.fragment_list = self.expected_dict
        f.write_csv(self.fragment_csv)
        output = []
        with open(self.fragment_csv, "r") as csvfile:
            csvreader = csv.reader(csvfile)
            for row in csvreader:
                output.append(row)
        self.assertListEqual(["fragment1", "8", "0", "11", "4", "40"], output[0])
        self.assertListEqual(["fragment2", "1", "1", "8", "8", "10"], output[1])

    def testReadsCsv(self):
        """Tests that the csv reads properly."""
        f = FragmentConfigHandler()
        f.read_csv(os.path.join("sample", "FragmentsTest.csv"))
        self.assertEqual(self.expected_dict2, f.fragment_list)
        with self.assertRaises(IOError):
            f.read_csv("not_a_file.csv")
        for file in ["FragmentsTestFail1.csv", "FragmentsTestFail2.csv"]:
            with self.assertRaises(IOError):
                f.read_csv(os.path.join("sample", file))
        with self.assertRaises(Exception):
            f.read_csv(os.path.join("sample", "dispersal_fine.tif"))


class TestFragmentCsvGeneration(unittest.TestCase):
    """
    Tests the generation of the fragment csv using the single function.
    """

    @classmethod
    def setUpClass(cls):
        """
        Defines the output csv.
        """
        cls.fragment_csv = os.path.join("output", "fragment_config1.csv")

    @classmethod
    def tearDownClass(cls):
        """
        Removes the generated csv file.
        """
        if os.path.exists(cls.fragment_csv):
            os.remove(cls.fragment_csv)

    def testGenerateFragmentCsv(self):
        """
        Tests that the fragments are generated correctly using the pipeline.
        """
        generate_fragment_csv(
            input_shapefile=os.path.join("sample", "shape_sample.shp"),
            input_raster=os.path.join("sample", "SA_sample_fine.tif"),
            output_csv=self.fragment_csv,
        )
        output = []
        with open(self.fragment_csv, "r") as csvfile:
            csvreader = csv.reader(csvfile)
            for row in csvreader:
                output.append(row)
        self.assertListEqual(["fragment2", "1", "1", "8", "8", "10"], output[1])
        self.assertListEqual(["fragment1", "8", "0", "11", "4", "40"], output[0])



"""
Tests the Merger module for combining the outputs of multiple simulations
"""

import os
import sys
import unittest

from setup_tests import setUpAll, tearDownAll

from pycoalescence import Merger, CoalescenceTree
from pycoalescence.sqlite_connection import fetch_table_from_sql, check_sql_table_exist


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestSimulationReadingViaMerger(unittest.TestCase):
    """
    Tests that simulation can successfully read tables from a database.
    """

    dbname = "output/mergers/combined1.db"

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)

    def testReadsSimulationParameters(self):
        """
        Tests that simulation parameters are correctly read from the database.
        """
        simulation_parameters = Merger()._read_simulation_parameters(input_simulation="sample/mergers/data_0_0.db")
        expected_parameters = [
            6,
            6,
            "output",
            0.5,
            4.0,
            4.0,
            1,
            0.1,
            10,
            1.0,
            1,
            0.0,
            200.0,
            "set",
            "sample/SA_sample_coarse.tif",
            35,
            41,
            11,
            14,
            1.0,
            "sample/SA_sample_fine.tif",
            13,
            13,
            0,
            0,
            "null",
            13,
            13,
            13,
            13,
            0,
            0,
            "none",
            "none",
            1,
            "normal",
            0.0,
            0.0,
            0,
            "closed",
            0,
            0.0,
            0.0,
            "none",
        ]
        self.assertListEqual(simulation_parameters, expected_parameters)

    def testReadsSpeciesList(self):
        """
        Tests that the species list object is correctly read from the database.
        """
        species_list = Merger()._read_species_list(input_simulation="sample/mergers/data_0_0.db")[0:5]
        expected_list = [
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0.0, 0, 0.0],
            [1, 0, 0, 0, 0, 0, 1, 0, 3761, 0, 0.712285394568117, 0, 0.0],
            [2, 0, 0, 0, 0, 0, 1, 0, 3762, 0, 0.113159547847957, 0, 0.0],
            [3, 0, 0, 0, 0, 0, 1, 0, 3763, 0, 0.95236636044045, 0, 0.0],
            [4, 0, 0, 0, 0, 0, 1, 0, 3764, 0, 0.679672305831754, 0, 0.0],
        ]
        for i, row in enumerate(expected_list):
            for j, each in enumerate(row):
                if j == 10 or j == 12:
                    self.assertAlmostEqual(each, species_list[i][j])
                else:
                    self.assertEqual(each, species_list[i][j])

    def testReadsSpeciesRichness(self):
        """
        Tests that the species richness object is correctly read from the database
        """
        species_richness = Merger()._read_species_richness(input_simulation="sample/mergers/data_0_0.db")
        expected_richness = [[0, 1, 3644], [1, 2, 3643], [2, 3, 3674], [3, 4, 3675], [4, 5, 3697], [5, 6, 3695]]
        for i, row in enumerate(expected_richness):
            self.assertListEqual(row, species_richness[i])

    def testReadsSpeciesAbundances(self):
        """
        Tests that the species abundances object is correctly read from the database
        """
        species_abundances = Merger()._read_species_abundances(input_simulation="sample/mergers/data_0_0.db")[-6:]
        expected_abundances = [
            [22029, 3690, 2, 6],
            [22030, 3691, 1, 6],
            [22031, 3692, 1, 6],
            [22032, 3693, 1, 6],
            [22033, 3694, 1, 6],
            [22034, 3695, 3, 6],
        ]
        for i, row in enumerate(expected_abundances):
            self.assertListEqual(row, species_abundances[i])

    def testReadsFragmentOctaves(self):
        """
        Tests that the fragment octaves object is correctly read from the database.
        """
        fragment_octaves = Merger()._read_fragment_octaves(input_simulation="sample/mergers/data_0_0.db")
        expected_octaves = [
            [0, "whole", 0, 3560],
            [1, "whole", 0, 3559],
            [2, "whole", 0, 3617],
            [3, "whole", 0, 3619],
            [4, "whole", 0, 3662],
            [5, "whole", 0, 3658],
        ]
        for i, row in enumerate(expected_octaves):
            self.assertListEqual(row, fragment_octaves[i])


class TestSimulationMerging(unittest.TestCase):
    """
    Tests that simulations can be successfully merged with a variety of tables and parameters.
    """

    @classmethod
    def setUpClass(cls):
        cls.dbname = "output/mergers/combined2.db"
        cls.dbname2 = "output/mergers/combined3.db"
        for each in [cls.dbname, cls.dbname2]:
            if os.path.exists(each):
                os.remove(each)
        cls.merger = Merger(database=cls.dbname)
        cls.merger.add_simulation("sample/mergers/data_0_0.db")
        cls.merger.add_simulation("sample/mergers/data_1_1.db")
        cls.merger.write()
        cls.merger2 = Merger(database=cls.dbname2)
        cls.merger2.add_simulations(["sample/mergers/data_2_2.db", "sample/mergers/data_3_3.db"])

    @classmethod
    def tearDownClass(cls):
        """
        Removes the output database
        """
        cls.merger.database.close()
        cls.merger.database = None
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)

    def testSetDatabase(self):
        """Tests that an error is raised if the output database already exists."""
        m = Merger()
        with self.assertRaises(IOError):
            m.set_database(self.dbname)

    def testMergerDeletion(self):
        """Tests that the merger deletes the database successfully."""
        m = Merger()
        m._open_database(filename=os.path.join("output", "mergertmp.db"))
        self.assertFalse(m.database is None)
        m.__del__()
        self.assertTrue(m.database is None)

    def testCreateTableErrors(self):
        """Tests that creating the simulation parameters raises the expected errors."""
        m = Merger()
        with self.assertRaises(IOError):
            m._create_simulation_parameters()
        with self.assertRaises(IOError):
            m._create_community_parameters()
        with self.assertRaises(IOError):
            m._create_metacommunity_parameters()
        with self.assertRaises(IOError):
            m._create_species_list()
        with self.assertRaises(IOError):
            m._create_species_locations()
        with self.assertRaises(IOError):
            m._create_species_abundances()
        with self.assertRaises(IOError):
            m._create_fragment_abundances()
        with self.assertRaises(IOError):
            m._create_fragment_octaves()
        with self.assertRaises(IOError):
            m._create_species_richness()
        with self.assertRaises(IOError):
            m._create_fragment_richness()
        with self.assertRaises(IOError):
            m._write_simulation_parameters()
        with self.assertRaises(IOError):
            m._write_species_list()
        with self.assertRaises(IOError):
            m._write_species_locations()
        with self.assertRaises(IOError):
            m._write_species_abundances()
        with self.assertRaises(IOError):
            m._write_fragment_abundances()
        with self.assertRaises(IOError):
            m._write_species_richness()
        with self.assertRaises(IOError):
            m._write_fragment_richness()
        with self.assertRaises(IOError):
            m._write_fragment_octaves()
        with self.assertRaises(IOError):
            m._write_community_parameters()
        with self.assertRaises(IOError):
            m._write_metacommunity_parameters()
        with self.assertRaises(ValueError):
            m._create_combined_species_richness()
        with self.assertRaises(ValueError):
            m._create_combined_fragment_richness()
        with self.assertRaises(ValueError):
            m._create_combined_fragment_octaves()

    def testSpeciesList(self):
        """
        Tests that the species list object is correctly calculates for the database
        """
        species_richness = fetch_table_from_sql(self.dbname, "SPECIES_LIST")
        check_richness = [x for x in species_richness if x[0] in [57236791, 57251923, 4, 5]]
        check_richness.sort(key=lambda x: x[0])
        expected_richness = [
            [4, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0.0, 0, 0.0, 2],
            [5, 0, 0, 0, 0, 0, 1, 0, 11328, 0, 0.712285394568117, 0, 0.0, 2],
            [57236791, 0, 8, 6, -1, 0, 0, 1, 0, 0, 0.59455712863121, 2, 0.0, 1],
            [57251923, 0, 7, 3, 0, 0, 0, 1, 0, 0, 0.603760047033245, 2, 0.0, 1],
        ]
        for i, row in enumerate(expected_richness):
            for j, each in enumerate(row):
                if j == 10 or j == 12:
                    self.assertAlmostEqual(each, check_richness[i][j])
                else:
                    self.assertEqual(each, check_richness[i][j])

    def testSpeciesRichness(self):
        """
        Tests that the landscape species richness is correctly calculated from the databases.
        """
        richness = [[0, 1, 7288], [1, 2, 7286], [2, 3, 7348], [3, 4, 7350], [4, 5, 7394], [5, 6, 7390]]
        self.assertListEqual(fetch_table_from_sql(self.dbname, "SPECIES_RICHNESS"), richness)

    def testSpeciesAbundances(self):
        """
        Tests that the species abundances are correctly calculated in the merged database
        """
        expected_abundances = [[5, 3697, 0, 1, 2], [1944632, 5090, 1, 1, 2], [485519191, 3695, 3, 6, 1]]
        real_abundances = fetch_table_from_sql(self.dbname, "SPECIES_ABUNDANCES")
        abundances = [x for x in real_abundances if x[0] == 1944632 or x[0] == 485519191 or x[0] == 5]
        if sys.version_info[0] >= 3:
            self.assertCountEqual(expected_abundances, abundances)
        else:  # pragma: no cover
            self.assertItemsEqual(expected_abundances, abundances)

    def testSimulationParameters(self):
        """
        Tests that the simulation parameters are correctly calculated in the merged database.
        """
        sim_pars = [
            [
                6,
                6,
                "output",
                0.5,
                4.0,
                4.0,
                1,
                0.1,
                10,
                1.0,
                1,
                0.0,
                200.0,
                "set",
                "sample/SA_sample_coarse.tif",
                35,
                41,
                11,
                14,
                1.0,
                "sample/SA_sample_fine.tif",
                13,
                13,
                0,
                0,
                "null",
                13,
                13,
                13,
                13,
                0,
                0,
                "none",
                "none",
                1,
                "normal",
                0.0,
                0.0,
                0,
                "closed",
                0,
                0.0,
                0.0,
                "none",
                1,
                "sample/mergers/data_0_0.db",
            ],
            [
                6,
                6,
                "output",
                0.5,
                4.0,
                4.0,
                2,
                0.1,
                10,
                1.0,
                1,
                0.0,
                200.0,
                "set",
                "sample/SA_sample_coarse.tif",
                35,
                41,
                11,
                14,
                1.0,
                "sample/SA_sample_fine.tif",
                13,
                13,
                0,
                0,
                "null",
                13,
                13,
                13,
                13,
                0,
                0,
                "none",
                "none",
                1,
                "normal",
                0.0,
                0.0,
                0,
                "closed",
                0,
                0.0,
                0.0,
                "none",
                2,
                "sample/mergers/data_1_1.db",
            ],
        ]
        actual = fetch_table_from_sql(self.dbname, "SIMULATION_PARAMETERS")
        for i, each in enumerate(sim_pars):
            self.assertListEqual(actual[i], each)

    def testSimulationParametersFetching(self):
        output_pars1 = {
            "coarse_map_file": "sample/SA_sample_coarse.tif",
            "coarse_map_scale": 1.0,
            "coarse_map_x": 35,
            "coarse_map_x_offset": 11,
            "coarse_map_y": 41,
            "coarse_map_y_offset": 14,
            "cutoff": 0.0,
            "deme": 1,
            "dispersal_map": "none",
            "dispersal_method": "normal",
            "dispersal_relative_cost": 1.0,
            "fine_map_file": "sample/SA_sample_fine.tif",
            "fine_map_x": 13,
            "fine_map_x_offset": 0,
            "fine_map_y": 13,
            "fine_map_y_offset": 0,
            "gen_since_historical": 200.0,
            "grid_x": 13,
            "grid_y": 13,
            "habitat_change_rate": 0.0,
            "landscape_type": "closed",
            "task": 6,
            "m_probability": 0.0,
            "max_speciation_gen": 0.0,
            "max_time": 10,
            "min_num_species": 1,
            "min_speciation_gen": 0.0,
            "output_dir": "output",
            "historical_coarse_map": "none",
            "historical_fine_map": "none",
            "protracted": 0,
            "sample_file": "null",
            "sample_size": 0.1,
            "sample_x": 13,
            "sample_x_offset": 0,
            "sample_y": 13,
            "sample_y_offset": 0,
            "seed": 6,
            "sigma": 4.0,
            "sim_complete": 1,
            "speciation_rate": 0.5,
            "tau": 4.0,
            "time_config_file": "set",
        }
        output_pars2 = {
            "coarse_map_file": "sample/SA_sample_coarse.tif",
            "coarse_map_scale": 1.0,
            "coarse_map_x": 35,
            "coarse_map_x_offset": 11,
            "coarse_map_y": 41,
            "coarse_map_y_offset": 14,
            "cutoff": 0.0,
            "deme": 2,
            "dispersal_map": "none",
            "dispersal_method": "normal",
            "dispersal_relative_cost": 1.0,
            "fine_map_file": "sample/SA_sample_fine.tif",
            "fine_map_x": 13,
            "fine_map_x_offset": 0,
            "fine_map_y": 13,
            "fine_map_y_offset": 0,
            "gen_since_historical": 200.0,
            "grid_x": 13,
            "grid_y": 13,
            "habitat_change_rate": 0.0,
            "landscape_type": "closed",
            "task": 6,
            "m_probability": 0.0,
            "max_speciation_gen": 0.0,
            "max_time": 10,
            "min_num_species": 1,
            "min_speciation_gen": 0.0,
            "output_dir": "output",
            "historical_coarse_map": "none",
            "historical_fine_map": "none",
            "protracted": 0,
            "sample_file": "null",
            "sample_size": 0.1,
            "sample_x": 13,
            "sample_x_offset": 0,
            "sample_y": 13,
            "sample_y_offset": 0,
            "seed": 6,
            "sigma": 4.0,
            "sim_complete": 1,
            "speciation_rate": 0.5,
            "tau": 4.0,
            "time_config_file": "set",
        }
        sim_pars1 = self.merger.get_simulation_parameters(guild=1)
        sim_pars2 = self.merger.get_simulation_parameters(guild=2)
        self.assertEqual(output_pars1, sim_pars1)
        self.assertEqual(output_pars2, sim_pars2)

    def testCommunityParameters(self):
        """Tests that the community parameters are correctly calculated in the merged database."""
        community_params = [
            [1, 0.5, 0.0, 0, 0],
            [2, 0.5, 0.5, 0, 0],
            [3, 0.6, 0.0, 0, 0],
            [4, 0.6, 0.5, 0, 0],
            [5, 0.7, 0.0, 0, 0],
            [6, 0.7, 0.5, 0, 0],
        ]
        actual = fetch_table_from_sql(self.dbname, "COMMUNITY_PARAMETERS")
        for i, each in enumerate(community_params):
            self.assertListEqual(each, actual[i])

    def testGenerateGuildTables(self):
        """Tests that the guild tables are generated properly."""
        self.assertFalse(check_sql_table_exist(self.merger2.file, "SPECIES_RICHNESS_GUILDS"))
        self.assertFalse(check_sql_table_exist(self.merger2.file, "FRAGMENT_RICHNESS_GUILDS"))
        self.assertFalse(check_sql_table_exist(self.merger2.file, "FRAGMENT_OCTAVES_GUILDS"))
        self.merger2.generate_guild_tables()
        self.assertTrue(check_sql_table_exist(self.merger2.file, "SPECIES_RICHNESS_GUILDS"))
        self.assertTrue(check_sql_table_exist(self.merger2.file, "FRAGMENT_RICHNESS_GUILDS"))
        self.assertTrue(check_sql_table_exist(self.merger2.file, "FRAGMENT_OCTAVES_GUILDS"))
        fragment_octaves = [
            [1, "fragment1", 0, 2, 1],
            [2, "fragment1", 1, 3, 1],
            [7, "fragment1", 0, 3, 1],
            [13, "fragment2", 0, 7, 1],
            [21, "fragment2", 1, 7, 1],
            [31, "fragment2", 0, 9, 1],
            [43, "fragment2", 1, 6, 1],
            [57, "whole", 0, 626, 1],
            [73, "whole", 1, 593, 1],
            [91, "whole", 2, 478, 1],
            [111, "whole", 3, 287, 1],
            [133, "whole", 4, 97, 1],
            [157, "whole", 0, 1035, 1],
            [183, "whole", 1, 985, 1],
            [211, "whole", 2, 651, 1],
            [241, "whole", 3, 238, 1],
            [273, "whole", 4, 44, 1],
            [4, "fragment1", 0, 6, 2],
            [5, "fragment1", 0, 10, 2],
            [6, "fragment2", 0, 10, 2],
            [14, "fragment2", 1, 6, 2],
            [22, "fragment2", 0, 18, 2],
            [32, "whole", 0, 708, 2],
            [44, "whole", 1, 563, 2],
            [58, "whole", 2, 494, 2],
            [74, "whole", 3, 275, 2],
            [92, "whole", 4, 97, 2],
            [112, "whole", 0, 1035, 2],
            [134, "whole", 1, 920, 2],
            [158, "whole", 2, 674, 2],
            [184, "whole", 3, 256, 2],
            [212, "whole", 0, 708, 2],
            [242, "whole", 1, 563, 2],
            [274, "whole", 2, 494, 2],
            [308, "whole", 3, 275, 2],
            [344, "whole", 4, 97, 2],
            [382, "whole", 0, 1035, 2],
            [422, "whole", 1, 920, 2],
            [464, "whole", 2, 674, 2],
            [508, "whole", 3, 256, 2],
        ]
        actual = fetch_table_from_sql(self.dbname2, "FRAGMENT_OCTAVES_GUILDS")
        for i, each in enumerate(fragment_octaves):
            self.assertListEqual(each, actual[i])
        richness = [[1, 1, 2095, 1], [2, 2, 2954, 1], [4, 1, 2147, 2], [5, 2, 2923, 2]]
        self.assertListEqual(fetch_table_from_sql(self.dbname2, "SPECIES_RICHNESS_GUILDS"), richness)


class TestMergerInheritance(unittest.TestCase):
    """
    Tests that the merger class correctly inherits from the CoalescenceTree class
    """

    @classmethod
    def setUpClass(cls):
        cls.dbname = "output/mergers/combined3.db"
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)
        cls.merger = Merger(database=cls.dbname)
        cls.merger.add_simulation("sample/mergers/data_0_0.db")
        cls.merger.add_simulation("sample/mergers/data_1_1.db")
        cls.merger.write()

    @classmethod
    def tearDownClass(cls):
        """
        Removes the output database
        """
        cls.merger.database.close()
        cls.merger.database = None
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)

    def testMergerGetSpeciesRichness(self):
        """
        Tests that species richness functions work properly.
        :return:
        """
        self.assertEqual(self.merger.get_species_richness(1), 7288)
        self.assertEqual(self.merger.get_species_richness(1), 7288)
        self.assertEqual(self.merger.get_species_richness(2), 7286)
        self.assertEqual(self.merger.get_species_richness(2), 7286)


class TestMergerAnalysis(unittest.TestCase):
    """
    Tests that the Merger can perform analysis in the same way as CoalescenceTree.
    """

    @classmethod
    def setUpClass(cls):
        cls.dbname = "output/mergers/combined4.db"
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)
        cls.merger = Merger(database=cls.dbname, logging_level=50)
        cls.merger.add_simulation("sample/mergers/data_0_0.db")
        cls.merger.add_simulation("sample/mergers/data_1_1.db")
        cls.merger.write()
        cls.merger.wipe_data()
        cls.merger.set_speciation_parameters(
            speciation_rates=[0.5, 0.6], record_spatial=False, record_fragments="sample/FragmentsTest.csv"
        )
        cls.merger.apply()
        cls.merger = Merger(database=cls.dbname, logging_level=50, expected=True)
        cls.merger.set_speciation_parameters(
            speciation_rates=[0.7, 0.8], record_spatial=False, record_fragments="sample/FragmentsTest.csv"
        )
        cls.merger.apply_incremental()
        cls.merger.output()
        cls.merger.calculate_fragment_richness()

    @classmethod
    def tearDownClass(cls):
        """
        Removes the output database
        """
        cls.merger.database.close()
        cls.merger.database = None
        if os.path.exists(cls.dbname):
            os.remove(cls.dbname)

    def testMergerRepeatSimAdding(self):
        """Tests that an error is raised when a repeat simulation is added."""
        ct = CoalescenceTree("sample/mergers/data_0_0.db")
        with self.assertRaises(ValueError):
            self.merger.add_simulation(ct)

    def testMergerAnalysisRichness(self):
        """
        Tests that speciation rates can be applied.
        """
        self.assertEqual(self.merger.get_species_richness(1), 7288)
        self.assertEqual(self.merger.get_species_richness(2), 7348)
        self.assertEqual(self.merger.get_species_richness(1), 7288)
        self.assertEqual(self.merger.get_species_richness(2), 7348)

    def testMergerAnalysisFragmentRichness(self):
        """
        Tests that the analysis can be performed.
        """
        self.assertEqual(self.merger.get_fragment_richness("fragment1", 1), 672)
        self.assertEqual(self.merger.get_fragment_richness("fragment2", 1), 1272)

    def testMergerOutput(self):
        """Tests that the Merger outputs correctly."""
        self.assertEqual(self.merger.get_species_richness(3), 7394)
        self.assertEqual(self.merger.get_species_richness(4), 7426)
        self.assertEqual(self.merger.get_fragment_richness("fragment2", 3), 1276)
        self.assertEqual(self.merger.get_fragment_richness("fragment2", 4), 1280)

    def testMergerAddedSimulations(self):
        """Tests that added simulations work as intended."""
        added_sims = self.merger.get_added_simulations()
        expected_sims = {"sample/mergers/data_0_0.db": 1, "sample/mergers/data_1_1.db": 2}
        self.assertEqual(added_sims, expected_sims)
        self.merger.simulation_list["sample/mergers/data_1_1.db"] = 4
        with self.assertRaises(ValueError):
            _ = self.merger.get_added_simulations()
        self.merger.simulation_list = expected_sims
